/*
Directory listing CGI program
By Steven Lawrance
(c) Steven Lawrance, 2000, 2002, 2004
(c) Bryant College, 1998-2000

Permission granted for release under the GPL in Q2 2000
by Bryant College.  Steven Lawrance can be contacted at
slawrance@yahoo.com.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "dirlist.h"
#include "globals.h"
#include "io.h"
#include "database.h"
#include "util.h"

/* Write to the file (encrypt if enabled) */
size_t WriteFile(const void *buffer, size_t size, size_t count, FILE *stream)
{
  /* If we have no encryption, then simply write the packet */
  if (!bEncryptEnable)
    return fwrite(buffer, size, count, stream);

  /* Switch based on the encryption method */
  switch (wEncryptMethod)
  {

#ifdef DIRLIST_USE_PACKETMANGLER
  case ENCRYPT_PACKETMANGLE_A :
    {
      /* Mangle this packet (version A; 8x8 grid-convert version) */
      BYTE chunk[12];
      DWORD dwBytesLeft = size * count;
      DWORD i, t, dwLevel;
      WORD wSum;
      BYTE cTemp;
      LPBYTE lpInput = (LPBYTE) buffer;
      while (dwBytesLeft >= 8)
      {
        /* Mangle these 8 bytes */
        wSum = (WORD) ftell(stream);
        ((LPWORD) chunk)[0] = ENCRYPT_PACKETMANGLE_A + wSum;
        for (i = 0; i < 8; i++)
          wSum += (WORD) (lpInput[i] * i);
        ((LPWORD) chunk)[1] = wSum;
        dwLevel = 1;
        for (i = 0; i < 8; i++) {
          chunk[4 + i] = 0;
          for (t = 0; t < 8; t++)
            chunk[4 + i] |= ((lpInput[t] & dwLevel) >> i) << t;
          dwLevel *= 2;
        }

        /* Twist the packet data */
        for (i = 0; i < 4; i++) {
          if (i % 2) {
            cTemp = chunk[4 + i];
            chunk[4 + i] = chunk[11 - i];
            chunk[11 - i] = cTemp + i;
          }
        }

        /* Confuse the data with the sum */
        wSum -= (WORD) ftell(stream);
        for (i = 0; i < 8; i++) {
          if (i % 3)
            chunk[4 + i] += (BYTE) (wSum >> 4);
          else
            chunk[4 + i] += (BYTE) wSum;
        }

        /* If the real writing function returned 12, then we're all set :-) */
        if (fwrite(chunk, 1, 12, stream) != 12)
          return 0;
        dwBytesLeft -= 8;
        lpInput = &(lpInput[8]);
      }
      if (dwBytesLeft)
      {
        /* Write a short packet */
        wSum = (WORD) ftell(stream);
        ((LPWORD) chunk)[0] = ENCRYPT_PACKETMANGLE_S + wSum;
        chunk[2] = (BYTE) dwBytesLeft;
        for (i = 0; i < dwBytesLeft; i++)
          chunk[3 + i] = lpInput[i] + (char) ((dwBytesLeft - i) * wSum);

        /* If the real writing function returned what we expected, then we're all set :-) */
        if (fwrite(chunk, 1, dwBytesLeft + 3, stream) != dwBytesLeft + 3)
          return 0;
      }

      /* Return the count requested */
      return count;
    }
#endif

  }

  /* Unknown encryption method! */
  infprintf("#Unknown encryption method, packet not written to file\n");
#ifndef DIRLIST_USE_PACKETMANGLER
  infprintf("#Try running configure with the --with-packetmangler=yes option and recompile.\n");
#endif
  return 0;
}

/* Read from the file (decrypt if needed) */
size_t ReadFile(void *buffer, size_t size, size_t count, FILE *stream)
{
  /* If we aren't using encryption, read the file normally; otherwise, determine encryption and decrypt */
  DWORD dwBytesLeft;
  LPBYTE lpOutput;
  WORD wMethod, wPos;
  if (!bEncryptEnable)
    return fread(buffer, size, count, stream);

  /* Process the encryption packets */
  dwBytesLeft = size * count;
  lpOutput = (LPBYTE) buffer;
  while (dwBytesLeft)
  {
    /* Read the encryption type */
    wPos = (WORD) ftell(stream);
    if (!ReadWORDNoEncrypt(stream, &wMethod))
      return 0;
    wMethod -= wPos;

    /* Switch based on the encryption type */
    switch (wMethod)
    {

#ifdef DIRLIST_USE_PACKETMANGLER
    case ENCRYPT_PACKETMANGLE_A :
      {
        /* Unmangle this packet (version A; 8x8 grid-convert version) */
        DWORD i, t, dwLevel;
        WORD wSum, wCheckSum;
        BYTE cTemp;
        BYTE chunk[8];

        /* If the requested buffer is not large enough, return an error */
        if (dwBytesLeft < 8) {
          infprintf("#Encrypted packet size error\n");
          return 0;
        }

        /* Get the sum */
        if (!ReadWORDNoEncrypt(stream, &wSum))
          return 0;

        /* Read the packet data */
        if (fread(chunk, 1, 8, stream) != 8)
          return 0;

        /* Clarify the data with the sum */
        wSum -= wPos;
        for (i = 0; i < 8; i++) {
          if (i % 3)
            chunk[i] -= (BYTE) (wSum >> 4);
          else
            chunk[i] -= (BYTE) wSum;
        }
        wSum += wPos;

        /* Untwist the packet data */
        for (i = 0; i < 4; i++) {
          if (i % 2) {
            cTemp = chunk[i];
            chunk[i] = chunk[7 - i] - i;
            chunk[7 - i] = cTemp;
          }
        }

        /* Unmangle these 8 bytes */
        dwLevel = 1;
        for (i = 0; i < 8; i++) {
          lpOutput[i] = 0;
          for (t = 0; t < 8; t++)
            lpOutput[i] |= ((chunk[t] & dwLevel) >> i) << t;
          dwLevel *= 2;
        }

        /* Compute the check-sum */
        wCheckSum = wPos;
        for (i = 0; i < 8; i++)
          wCheckSum += (WORD) (lpOutput[i] * i);

        /* Compare the check-sum */
        if (wCheckSum != wSum) {
          infprintf("#Invalid checksum in encrypted packet\n");
          return 0;
        }

        /* Decrease the bytes left by 8 */
        dwBytesLeft -= 8;
        lpOutput = &(lpOutput[8]);
        break;
      }

    case ENCRYPT_PACKETMANGLE_S :
      {
        /* Read the packet length */
        DWORD i;
        BYTE nBytes;
        if (!fread(&nBytes, 1, 1, stream))
          return 0;

        /* If the length is greater than the bytes requested, then return an error */
        if ((nBytes > dwBytesLeft) || (nBytes > 7)) {
          infprintf("#Encrypted packet size error\n");
          return 0;
        }

        /* Read in the encrypted packet */
        if (fread(lpOutput, 1, nBytes, stream) != nBytes)
          return 0;

        /* Decrypt the data */
        for (i = 0; i < nBytes; i++) {
          lpOutput[i] -= (char) (((DWORD) nBytes - i) * wPos);
        }

        /* Decrease the number of bytes left */
        dwBytesLeft -= nBytes;
        lpOutput = &(lpOutput[nBytes]);
        break;
      }
#endif

    default :
      {
        /* Unknown encryption method! */
        infprintf("#Unknown encryption method, packet not read from file.\n");
#ifndef DIRLIST_USE_PACKETMANGLER
        infprintf("#Try running configure with the --with-packetmangler=yes option and recompile.\n");
#endif
        return 0;
      }
    }
  }

  /* Return the count */
  return count;
}

int SeekFile(FILE *stream, long offset, int origin)
{
  /* If encryption is disabled, the origin is absolute, or we're seeking backwards, then simply seek */
  DWORD dwBytesLeft;
  WORD wMethod, wPos;
  if (!bEncryptEnable || (origin == SEEK_SET) || (origin == SEEK_END) || (offset < 0))
    return fseek(stream, offset, origin);

  /* Seek ahead the requested number of virtual bytes */
  dwBytesLeft = (DWORD) offset;
  while (dwBytesLeft)
  {
    /* Read the encrypted packet header */
    wPos = (WORD) ftell(stream);
    if (!ReadWORDNoEncrypt(stream, &wMethod))
      return -1;
    wMethod -= wPos;

    /* Switch based on the encryption type */
    switch (wMethod)
    {

#ifdef DIRLIST_USE_PACKETMANGLER
    case ENCRYPT_PACKETMANGLE_A :
      {
        /* If the requested seek is not large enough, return an error */
        if (dwBytesLeft < 8)
          return -1;

        /* Seek ahead 10 bytes (8 virtually) */
        if (fseek(stream, 10, SEEK_CUR))
          return -1;

        /* Decrease the bytes left by 8 */
        dwBytesLeft -= 8;
        break;
      }

    case ENCRYPT_PACKETMANGLE_S :
      {
        /* Read the packet length */
        BYTE nBytes;
        if (!fread(&nBytes, 1, 1, stream))
          return -1;

        /* Seek ahead by the number of data bytes */
        if (fseek(stream, (long) nBytes, SEEK_CUR))
          return -1;

        /* Decrease the number of bytes left */
        dwBytesLeft -= nBytes;
        break;
      }
#endif

    default :
      {
        /* Unknown encryption method!  Simply seek ahead by the requested number */
        printf("#Unknown encryption method detected while seeking\n");
#ifndef DIRLIST_USE_PACKETMANGLER
        infprintf("#Try running configure with the --with-packetmangler=yes option and recompile.\n");
#endif
        return fseek(stream, (long) dwBytesLeft, SEEK_CUR);
      }
    }
  }

  /* Return a success */
  return 0;
}

static inline const DWORD SwabWORD(WORD val)
{
  return ((val & 0x00FF) << 8) |
         ((val & 0xFF00) >> 8);
}

static inline const DWORD SwabDWORD(DWORD val)
{
  return ((val & 0x000000FF) << 24) |
         ((val & 0x0000FF00) << 8)  |
         ((val & 0x00FF0000) >> 8) |
         ((val & 0xFF000000) >> 24);
}

BOOL ReadWORDNoEncrypt(FILE *index, WORD *value)
{
  /* Read a DWORD from the file and perform any necessary conversions */
  if (fread(value, sizeof(WORD), 1, index) < 1)
    return FALSE;

  /* If we are performing conversions, then convert now */
  if (bConvertEndian)
    *value = SwabWORD(*value);

  /* Return success */
  return TRUE;
}

BOOL ReadWORD(FILE *index, WORD *value)
{
  /* Read a DWORD from the file and perform any necessary conversions */
  if (ReadFile(value, sizeof(WORD), 1, index) < 1)
    return FALSE;

  /* If we are performing conversions, then convert now */
  if (bConvertEndian)
    *value = SwabWORD(*value);

  /* Return success */
  return TRUE;
}

BOOL ReadDWORDNoEncrypt(FILE *index, DWORD *value)
{
  /* Read a DWORD from the file and perform any necessary conversions */
  if (fread(value, sizeof(DWORD), 1, index) < 1)
    return FALSE;

  /* If we are performing conversions, then convert now */
  if (bConvertEndian)
    *value = SwabDWORD(*value);

  /* Return success */
  return TRUE;
}

BOOL ReadDWORD(FILE *index, DWORD *value)
{
  /* Read a DWORD from the file and perform any necessary conversions */
  if (ReadFile(value, sizeof(DWORD), 1, index) < 1)
    return FALSE;

  /* If we are performing conversions, then convert now */
  if (bConvertEndian)
    *value = SwabDWORD(*value);

  /* Return success */
  return TRUE;
}

/* Read the entries section of a sort table into an array */
LPDWORD ReadSortEntriesArray(FILE *index, LPDIRLIST_SORTTABLE lpTable)
{
  /* Read the sort table into an unobtrusive array */
  LPDWORD lpdwArray = (LPDWORD) malloc(sizeof(DWORD) * lpTable->nEntries);
  DWORD dwOldFilePos = (DWORD) ftell(index);
  DWORD i;
  BOOL bExit = FALSE;

  /* Seek to the sort table position after saving the current position */
  fseek(index, (long) lpTable->dwFilePos, SEEK_SET);

  /* Loop through each entry (or until a read fails, whichever comes first) */
  for (i = 0; (i < lpTable->nEntries) && !bExit; i++)
  {
    /* If the read failed, then set bExit */
    if (!ReadDWORDNoEncrypt(index, &(lpdwArray[i])))
      bExit = TRUE;
  }

  /* Seek to the old file position (we want to be polite ;-) */
  fseek(index, (long) dwOldFilePos, SEEK_SET);

  /* If we came across a read error, then free the array and return NULL */
  if (bExit) {
    free(lpdwArray);
    return NULL;
  }

  /* Return the array :-)! */
  return lpdwArray;
}

/* Check the bad template */
BOOL CheckBadTemplate(FILE *pFile, int badTemplateNumber)
{
/* Check this page against Bryant's template; if there is a match,
  then this page is invalid */
  char testBad[80], testHTML[80];
  FILE *badTemplate;
  BOOL resume = TRUE;
  int nBytes;
  BOOL match = TRUE;
  sprintf(testBad, "dirlist_html%cbadtemplate%i.html", cDirDelimiter, badTemplateNumber);
  badTemplate = fopen(testBad, "r");

  if (!badTemplate) {
    return TRUE;
  }

  do {
    memset(testBad, 0, 80);
    memset(testHTML, 0, 80);

    if (fgets(testBad, 80, badTemplate)) {
      nBytes = strlen(testBad);
      testBad[nBytes - 1] = '\0';
      nBytes--;
      if (testBad) {
        if (fgets(testHTML, 80, pFile)) {
          /* Compare the two strings */
          if (strncmp(testBad, testHTML, nBytes) != 0)
            match = FALSE;
        } else
          resume = FALSE;
      }
    } else
      resume = FALSE;
  } while (resume && match);
  /* Close our files */
  fclose(badTemplate);

  /* Return our success/failure */
  if (match)
    return FALSE;
  return TRUE;
}

void WriteString(FILE *index, LPCTSTR str, DWORD len)
{
  /* Write a string using the wacky string storage schema we're using ;-) */
  fwrite(&len, sizeof(DWORD), 1, index);
  WriteFile(str, len + 1, 1, index);
}

BOOL WriteIndexEntry(FILE *index, void *lpData, UCHAR dwType, BOOL bEmitStatus)
{
  /* Type 0 index entry (Header data) */
  if (dwType == 0)
  {
    /* Variables */
    LPDIRLIST_HEADER lpHeader = (LPDIRLIST_HEADER) lpData;
    BOOL bEncrypt = bEncryptEnable;
    DWORD lWebDocsDir = strlen(lpHeader->lpszWebDocsDir);
    DWORD lBaseDir = strlen(lpHeader->lpszBaseDir);
    DWORD lAuditEMail = strlen(lpHeader->lpszAuditEMail);
    DWORD lEMailDomain = strlen(lpHeader->lpszEMailDomain);
    DWORD lDefaultURL = strlen(lpHeader->lpszDefaultURL);
    DWORD len;

    /* Write the index entry header */
    fwrite("@0", 1, 2, index);

    /* Write the total length of the entry */
    len = sizeof(DWORD) * 12 + lWebDocsDir + lBaseDir + 2 + ((lpHeader->dwFlags & HEADER_SEQUENCENUMBERS)? sizeof(DWORD): 0) + ((lpHeader->dwFlags & HEADER_ENFORCELOCKS)? sizeof(DWORD): 0) + ((lpHeader->dwFlags & HEADER_AUDITEMAIL)? sizeof(DWORD) + lAuditEMail + 1: 0) + ((lpHeader->dwFlags & HEADER_EMAILDOMAIN)? sizeof(DWORD) + lEMailDomain + 1: 0) + ((lpHeader->dwFlags & HEADER_DEFAULTURL)? sizeof(DWORD) + lDefaultURL + 1: 0);
    fwrite(&len, sizeof(DWORD), 1, index);

    /* Write the version information */
    fwrite(&(lpHeader->dwVersion), sizeof(DWORD), 1, index);
    fwrite(&(lpHeader->dwLowestVersion), sizeof(DWORD), 1, index);

    /* Write the statistical data */
    fwrite(&(lpHeader->nEntryTypes), sizeof(DWORD), 1, index);
    fwrite(&(lpHeader->nType1Entries), sizeof(DWORD), 1, index);
    fwrite(&(lpHeader->nType2Entries), sizeof(DWORD), 1, index);
    fwrite(&(lpHeader->nType3Entries), sizeof(DWORD), 1, index);
    fwrite(&(lpHeader->nType4Entries), sizeof(DWORD), 1, index);
    fwrite(&(lpHeader->nType5Entries), sizeof(DWORD), 1, index);
    fwrite(&(lpHeader->nType6Entries), sizeof(DWORD), 1, index);

    /* Write the flags and the directory locations */
    fwrite(&(lpHeader->dwFlags), sizeof(DWORD), 1, index);
    bEncryptEnable = FALSE;
    WriteString(index, lpHeader->lpszWebDocsDir, lWebDocsDir);
    WriteString(index, lpHeader->lpszBaseDir, lBaseDir);

    /* If the sequence numbers flag is set, then write the sequence number */
    if (lpHeader->dwFlags & HEADER_SEQUENCENUMBERS)
      fwrite(&(lpHeader->dwSequenceNumber), sizeof(DWORD), 1, index);

    /* If we are supposed to enforce locks, then write the lock information */
    if (lpHeader->dwFlags & HEADER_ENFORCELOCKS)
      fwrite(&(lpHeader->dwFilePosLock), sizeof(DWORD), 1, index);

    /* If we have an audit e-mail address, then write the e-mail address */
    if (lpHeader->dwFlags & HEADER_AUDITEMAIL)
      WriteString(index, lpHeader->lpszAuditEMail, lAuditEMail);

    /* If we have an e-mail domain, then write the e-mail domain */
    if (lpHeader->dwFlags & HEADER_EMAILDOMAIN)
      WriteString(index, lpHeader->lpszEMailDomain, lEMailDomain);

    /* If we have a default URL, then write it */
    if (lpHeader->dwFlags & HEADER_DEFAULTURL)
      WriteString(index, lpHeader->lpszDefaultURL, lDefaultURL);
    bEncryptEnable = bEncrypt;

    /* We did it :-) */
    return TRUE;
  }

  /* Type 1 index entry */
  if (dwType == 1) {
    infprintf("#Version 1 entries are no longer supported\n");
    return FALSE;
  }

  /* Type 2 index entry */
  if (dwType == 2)
  {
    /* Compute the string lengths */
    LPDIRLIST_ENTRY lpDirEntry = (LPDIRLIST_ENTRY) lpData;
    DWORD lUserName  = strlen(lpDirEntry->lpszUserName);
    DWORD lFirstName = strlen(lpDirEntry->lpszFirstName);
    DWORD lLastName  = strlen(lpDirEntry->lpszLastName);
    DWORD lWebSite   = strlen(lpDirEntry->lpszWebSite);
    DWORD lEMail     = strlen(lpDirEntry->lpszEMail);
    DWORD lTelephone = strlen(lpDirEntry->lpszTelephone);
    DWORD *lValueNames = (DWORD *) malloc(sizeof(DWORD) * lpDirEntry->nClassDataValues);
    DWORD len = 0, i, deletedClasses = 0;

    /* Write the index entry, version 2 */
    fwrite("@2", 1, 2, index);

    /* Determine the lengths of the class data strings to write */
    if (lpDirEntry->nClassDataValues)
    {
      /* Space for the nClassDataValues variable */
      len = sizeof(DWORD);

      /* Loop through this entry's data */
      for (i = 0; i < lpDirEntry->nClassDataValues; i++) {
        if (!lpDirEntry->lpClassData) {
          infprintf("#In WriteIndexEntry(), lpDirEntry->lpClassData is NULL in index %u!", i); fflush(stdout);
          return FALSE;
        }
        if (lpDirEntry->lpClassData[i].idClass) {
          if (lpDirEntry->lpClassData[i].dwFlags & CLASSDATA_OTHER)
          {
            /* "None of the above"-type string, write string value */
            lValueNames[i] = strlen(lpDirEntry->lpClassData[i].lpszValue);
            len += lValueNames[i] + sizeof(DWORD) * 3 + 1;
          } else
          {
            /* Class-defined string, write ordinal value of string instead */
            lValueNames[i] = 0;
            len += sizeof(DWORD) * 3;
          }
        } else
        {
          /* Deleted entry (class value = 0), do not write */
          lValueNames[i] = 0;
          deletedClasses++;
        }
      }

      /* Set the FLAGS_CLASSDATA if we have class data */
      if (lpDirEntry->nClassDataValues - deletedClasses)
        lpDirEntry->dwFlags |= FLAGS_CLASSDATA;
      else {
        lpDirEntry->dwFlags &= ~FLAGS_CLASSDATA;
        len -= sizeof(DWORD);
        free(lpDirEntry->lpClassData);
        lpDirEntry->lpClassData = NULL;
      }
    } else
    {
      /* Ensure the FLAGS_CLASSDATA flag is not set */
      lpDirEntry->dwFlags &= ~FLAGS_CLASSDATA;
    }

    /* Write the total length of the entry */
    len += lUserName + lFirstName + lLastName + lWebSite + lEMail + lTelephone + sizeof(DWORD) * 7 + 6;
    WriteFile(&len, sizeof(DWORD), 1, index);

    /* Write the flags */
    WriteFile(&(lpDirEntry->dwFlags), sizeof(DWORD), 1, index);

    /* Write the strings */
    WriteString(index, lpDirEntry->lpszUserName, lUserName);
    WriteString(index, lpDirEntry->lpszFirstName, lFirstName);
    WriteString(index, lpDirEntry->lpszLastName, lLastName);
    WriteString(index, lpDirEntry->lpszWebSite, lWebSite);
    WriteString(index, lpDirEntry->lpszEMail, lEMail);
    WriteString(index, lpDirEntry->lpszTelephone, lTelephone);

    /* Write any class data that we may have */
    if (lpDirEntry->nClassDataValues - deletedClasses)
    {
      /* Write the nClassDataValues variable */
      lpDirEntry->nClassDataValues -= deletedClasses;
      WriteFile(&(lpDirEntry->nClassDataValues), sizeof(DWORD), 1, index);

      /* Loop through this entry's data */
      for (i = 0; i < lpDirEntry->nClassDataValues + deletedClasses; i++) {
        if (lpDirEntry->lpClassData[i].idClass)
        {
          /* Write the class data header */
          WriteFile(&(lpDirEntry->lpClassData[i].idClass), sizeof(DWORD), 1, index);
          WriteFile(&(lpDirEntry->lpClassData[i].dwFlags), sizeof(DWORD), 1, index);

          /* Write the value of the class data */
          if (lpDirEntry->lpClassData[i].dwFlags & CLASSDATA_OTHER)
          {
            /* "None of the above"-type string, write string value */
            WriteString(index, lpDirEntry->lpClassData[i].lpszValue, lValueNames[i]);
          } else
          {
            /* Class-defined string, write ordinal value of string instead */
            WriteFile(&(lpDirEntry->lpClassData[i].dwValue), sizeof(DWORD), 1, index);
          }
        }
      }
    }
    free(lValueNames);

    /* Inform the user of our action if we're supposed to */
    if (bEmitStatus) {
      DWORD origSafeStringsLen = safeStringsLen;
      infprintf("#User \"%s\" written to database\n",
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpDirEntry->lpszUserName, STR_NUL_TERM));
      safeStringsLen = origSafeStringsLen;
    }
    return TRUE;
  }

  /* Type 3 class information */
  if (dwType == 3)
  {
    /* Compute the string lengths */
    LPDIRLIST_CLASS lpClass = (LPDIRLIST_CLASS) lpData;
    DWORD lClassName = strlen(lpClass->lpszName);
    DWORD *lValueNames = (DWORD *) malloc(sizeof(DWORD) * lpClass->nValues);
    DWORD len = lClassName, i;
    DWORD dwReserved = 0;

    /* Write the index entry id of 3 */
    fwrite("@3", 1, 2, index);

    /* Collect the string lengths of the value names */
    for (i = 0; i < lpClass->nValues; i++) {
      if (lpClass->lpValues[i].lpszValue) {
        lValueNames[i] = strlen(lpClass->lpValues[i].lpszValue);
        len += lValueNames[i] + sizeof(DWORD) * 2 + 1;
      } else
        lValueNames[i] = 0;
    }

    /* Write the total length of the entry */
    len += sizeof(DWORD) * 6 + 1;
    WriteFile(&len, sizeof(DWORD), 1, index);

    /* Write the class ID, name, number of values, and zero (for no extended class data) */
    WriteFile(&(lpClass->idClass), sizeof(DWORD), 1, index);
    WriteString(index, lpClass->lpszName, lClassName);
    WriteFile(&(lpClass->nValues), sizeof(DWORD), 1, index);
    WriteFile(&(lpClass->dwFlags), sizeof(DWORD), 1, index);
    WriteFile(&(lpClass->idDefaultValue), sizeof(DWORD), 1, index);
    WriteFile(&(dwReserved), sizeof(DWORD), 1, index);

    /* Write the values that have non-NULL strings (NULL strings are "deleted") */
    for (i = 0; i < lpClass->nValues; i++) {
      if (lpClass->lpValues[i].lpszValue) {
        WriteFile(&(lpClass->lpValues[i].idValue), sizeof(DWORD), 1, index);
        WriteString(index, lpClass->lpValues[i].lpszValue, lValueNames[i]);
      }
    }
    free(lValueNames);

    /* Inform the user of our action if we're supposed to */
    if (bEmitStatus) {
      DWORD origSafeStringsLen = safeStringsLen;
      infprintf("#Class (and its values) \"%s\" written to database\n",
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpClass->lpszName, STR_NUL_TERM));
      safeStringsLen = origSafeStringsLen;
    }
    return TRUE;
  }

  /* Type 4 sort table */
  if (dwType == 4)
  {
    /* Compute the length */
    LPDIRLIST_SORTTABLE lpTable = (LPDIRLIST_SORTTABLE) lpData;
    LPDIRLIST_SORTENTRY lpCurrent;
    DWORD lOrder = strlen(lpTable->lpszOrder);
    DWORD len = lpTable->nEntries * sizeof(DWORD) + 3 * sizeof(DWORD) + lOrder + 1;
    BOOL bEncrypt = bEncryptEnable;

    /* Write the index entry id of 4 */
    fwrite("@4", 1, 2, index);

    /* Write the total length of the entry */
    fwrite(&len, sizeof(DWORD), 1, index);

    /* Write the sort table ID, number of entries, and order string */
    fwrite(&(lpTable->idSort), sizeof(DWORD), 1, index);
    fwrite(&(lpTable->nEntries), sizeof(DWORD), 1, index);
    bEncryptEnable = FALSE;
    WriteString(index, lpTable->lpszOrder, lOrder);
    bEncryptEnable = bEncrypt;

    /* Write the locations to the file */
    lpCurrent = lpTable->lpSortEntries;
    while (lpCurrent) {
      fwrite(&(lpCurrent->lpDirEntry->dwFilePos), sizeof(DWORD), 1, index);
      lpCurrent = (LPDIRLIST_SORTENTRY) lpCurrent->next;
    }

    /* Inform the user of our action if we're supposed to */
    if (bEmitStatus) {
      infprintf("#Sort table %u written to database\n", lpTable->idSort);
    }
    return TRUE;
  }

  /* Type 5 user information */
  if (dwType == 5)
  {
    /* Compute the length */
    LPDIRLIST_USER lpUser = (LPDIRLIST_USER) lpData;
    DWORD lName = strlen(lpUser->lpszName);
    DWORD lDescription = strlen(lpUser->lpszDescription);
    DWORD lPassword = strlen(lpUser->lpszPassword);
    DWORD len = lName + lDescription + lPassword + 3 * (sizeof(DWORD) + 1) + 5 * sizeof(DWORD) + 2 * sizeof(BOOL);

    /* Write the index entry id of 5 */
    fwrite("@5", 1, 2, index);

    /* Write the total length of the entry */
    WriteFile(&len, sizeof(DWORD), 1, index);

    /* Write the standard information */
    WriteFile(&(lpUser->idUser), sizeof(DWORD), 1, index);
    WriteString(index, lpUser->lpszName, lName);
    WriteString(index, lpUser->lpszPassword, lPassword);
    WriteString(index, lpUser->lpszDescription, lDescription);

    /* Write the user level */
    WriteFile(&(lpUser->dwUserLevel), sizeof(DWORD), 1, index);

    /* Write the permitted operations */
    WriteFile(&(lpUser->dwOperations), sizeof(DWORD), 1, index);
    WriteFile(&(lpUser->bAdminMode), sizeof(BOOL), 1, index);
    WriteFile(&(lpUser->dwSearchLevel), sizeof(DWORD), 1, index);

    /* Write the auditing information */
    WriteFile(&(lpUser->bAudit), sizeof(BOOL), 1, index);
    WriteFile(&(lpUser->dwAuditOperations), sizeof(DWORD), 1, index);
    return TRUE;
  }

  /* Type 6 locking information */
  if (dwType == 6)
  {
    /* Compute the length */
    LPDIRLIST_LOCK lpLock = (LPDIRLIST_LOCK) lpData;
    DWORD len = 3 * sizeof(DWORD) + sizeof(BOOL) + sizeof(time_t);

    /* Write the index entry id of 6 */
    fwrite("@6", 1, 2, index);

    /* Write the total length of the entry */
    WriteFile(&len, sizeof(DWORD), 1, index);

    /* Write the lock information */
    WriteFile(&(lpLock->dwFilePosUser), sizeof(DWORD), 1, index);
    WriteFile(&(lpLock->dwOperations), sizeof(DWORD), 1, index);
    WriteFile(&(lpLock->bAdminMode), sizeof(BOOL), 1, index);
    WriteFile(&(lpLock->tLockTime), sizeof(time_t), 1, index);
    WriteFile(&(lpLock->dwLockDuration), sizeof(DWORD), 1, index);
    return TRUE;
  }

  /* No handler for the requested data type!  Return FALSE */
  return FALSE;
}

int ReadString(FILE *index, LPTSTR *str)
{
  /* Read a string from the file */
  DWORD len;
  if (!ReadDWORDNoEncrypt(index, &len))
    return FALSE;
  *str = (LPTSTR) malloc(len + 1);
  if (ReadFile(*str, len + 1, 1, index) < 1) {
    free(*str);
    return FALSE;
  }
  (*str)[len] = 0;

  /* Return the number of bytes the string takes up in the file */
  return len + 1 + (int) sizeof(DWORD);
}

int ReadIndexEntry(FILE *index, LPBYTE *lpData, LPDIRLIST_CLASS lpClassList)
{
  /* Read the index entry, skipping over null bytes (added in release 3) */
  char type[3];
  DWORD len;
  if (fread(type, 1, 1, index) < 1)
    return -3;
  while (!type[0]) {
    if (fread(type, 1, 1, index) < 1)
      return -3;
  }
  if (fread(&(type[1]), 1, 1, index) < 1)
    return -2;
  type[2] = 0;

  /* Jump point (single mode; added in release 3) */
  if (strcmp(type, "@J") == 0)
  {
    /* Jump to the given location */
    DWORD dwFilePosJump, dwFilePosCurrent;
    int iRetVal;
    if (!ReadDWORD(index, &dwFilePosJump))
      return -2;
    dwFilePosCurrent = (DWORD) ftell(index) + 2 + sizeof(DWORD);
    SeekFile(index, (long) dwFilePosJump, SEEK_SET);

    /* Call ourselves to read the entry */
    iRetVal = ReadIndexEntry(index, lpData, lpClassList);

    /* Return to the previous position and return the return value */
    SeekFile(index, (long) dwFilePosCurrent, SEEK_SET);
    return iRetVal;
  }

  /* Jump point targets; skip ahead of them */
  if (strcmp(type, "@T") == 0)
  {
    /* Read the length and skip past it */
    DWORD v2len;
    if (!ReadDWORD(index, &v2len))
      return -2;
    SeekFile(index, (long) v2len, SEEK_CUR);

    /* Return -3 for a jump point target */
    return -3;
  }

  /* Type 0 index entry (Header data) */
  if (strcmp(type, "@0") == 0)
  {
    /* Allocate the memory */
    LPDIRLIST_HEADER lpHeader = (LPDIRLIST_HEADER) malloc(sizeof(DIRLIST_HEADER));
    BOOL bEncrypt = bEncryptEnable;
    DWORD v2len;
    memset(lpHeader, (char) NULL, sizeof(DIRLIST_HEADER));
    *lpData = (LPBYTE) lpHeader;
    if (!ReadDWORDNoEncrypt(index, &v2len)) {
      free(lpHeader);
      return -2;
    }

    /* Read version information */
    if (!ReadDWORDNoEncrypt(index, &(lpHeader->dwVersion))) {
      free(lpHeader);
      return -2;
    }

    /* If the two highest-order bytes have data, then we should try swapping the byteorder */
    if (lpHeader->dwVersion & 0xFFFF0000) {
      DWORD oldVer = lpHeader->dwVersion;
      lpHeader->dwVersion = SwabDWORD(lpHeader->dwVersion);

      /* If the swapped byteorder worked, then remind ourselves to swap all read-in DWORDs */
      if (!(lpHeader->dwVersion & 0xFFFF0000)) {
#ifdef WORDS_BIGENDIAN
        infprintf("#Importing from little-endian index file\n");
#else
        infprintf("#Importing from big-endian index file\n");
#endif
        bConvertEndian = TRUE;

        /* Fix up v2len */
        v2len = SwabDWORD(v2len);
      }

      /* Swapping the byteorder didn't create a sane version number, so use the original */
      else
        lpHeader->dwVersion = oldVer;
    }
    v2len -= sizeof(DWORD);

    /* Read in the lowest minor version that this file supports */
    if (!ReadDWORDNoEncrypt(index, &(lpHeader->dwLowestVersion))) {
      free(lpHeader);
      return -2;
    }
    v2len -= sizeof(DWORD);

    /* Read the statistical data */
    if (!ReadDWORDNoEncrypt(index, &(lpHeader->nEntryTypes))) {
      free(lpHeader);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (!ReadDWORDNoEncrypt(index, &(lpHeader->nType1Entries))) {
      free(lpHeader);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (!ReadDWORDNoEncrypt(index, &(lpHeader->nType2Entries))) {
      free(lpHeader);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (!ReadDWORDNoEncrypt(index, &(lpHeader->nType3Entries))) {
      free(lpHeader);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (!ReadDWORDNoEncrypt(index, &(lpHeader->nType4Entries))) {
      free(lpHeader);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (lpHeader->nEntryTypes >= 6) {
      if (!ReadDWORDNoEncrypt(index, &(lpHeader->nType5Entries))) {
        free(lpHeader);
        return -2;
      }
      v2len -= sizeof(DWORD);
      if (!ReadDWORDNoEncrypt(index, &(lpHeader->nType6Entries))) {
        free(lpHeader);
        return -2;
      }
      v2len -= sizeof(DWORD);
    }

    /* If more entry types exist, then skip ahead of them */
    if ((lpHeader->nEntryTypes > 6) && v2len) {
      SeekFile(index, (lpHeader->nEntryTypes - 6) * sizeof(DWORD), SEEK_CUR);
      v2len -= (lpHeader->nEntryTypes - 6) * sizeof(DWORD);
    }

    /* If more bytes exist, then read the flags, web docs directory, and base directory */
    if (v2len) {
      DWORD len;
      if (!ReadDWORDNoEncrypt(index, &(lpHeader->dwFlags))) {
        free(lpHeader);
        return -2;
      }
      v2len -= sizeof(DWORD);
      bEncryptEnable = FALSE;
      if (!(len = ReadString(index, &(lpHeader->lpszWebDocsDir)))) {
        free(lpHeader);
        return -2;
      }
      v2len -= len;
      if (!(len = ReadString(index, &(lpHeader->lpszBaseDir)))) {
        free(lpHeader->lpszWebDocsDir);
        free(lpHeader);
        return -2;
      }
      v2len -= len;

      /* If the sequence number flag is set, then read the sequence number */
      if (v2len && (lpHeader->dwFlags & HEADER_SEQUENCENUMBERS)) {
        if (!ReadDWORDNoEncrypt(index, &(lpHeader->dwSequenceNumber))) {
          free(lpHeader->lpszWebDocsDir);
          free(lpHeader->lpszBaseDir);
          free(lpHeader);
          return -2;
        }
        v2len -= sizeof(DWORD);
      }

      /* If the lock enforcement flag is set, then read the lock information file position */
      if (v2len && (lpHeader->dwFlags & HEADER_ENFORCELOCKS)) {
        if (!ReadDWORDNoEncrypt(index, &(lpHeader->dwFilePosLock))) {
          free(lpHeader->lpszWebDocsDir);
          free(lpHeader->lpszBaseDir);
          free(lpHeader);
          return -2;
        }
        v2len -= sizeof(DWORD);
      }

      /* If the audit e-mail flag is set, then read the audit e-mail address */
      if (v2len && (lpHeader->dwFlags & HEADER_AUDITEMAIL)) {
        if (!(len = ReadString(index, &(lpHeader->lpszAuditEMail)))) {
          free(lpHeader->lpszWebDocsDir);
          free(lpHeader->lpszBaseDir);
          free(lpHeader);
          return -2;
        }
        SafeEMailAddress(lpHeader->lpszAuditEMail);
        v2len -= len;
      }

      /* If the e-mail domain flag was set, then read it */
      if (v2len && (lpHeader->dwFlags & HEADER_EMAILDOMAIN)) {
        if (!(len = ReadString(index, &(lpHeader->lpszEMailDomain)))) {
          free(lpHeader->lpszAuditEMail);
          free(lpHeader->lpszWebDocsDir);
          free(lpHeader->lpszBaseDir);
          free(lpHeader);
          return -2;
        }
        SafeDomainName(lpHeader->lpszEMailDomain);
        v2len -= len;
      }

      /* If the default URL flag was set, then read it */
      if (v2len && (lpHeader->dwFlags & HEADER_DEFAULTURL)) {
        if (!(len = ReadString(index, &(lpHeader->lpszDefaultURL)))) {
          free(lpHeader->lpszEMailDomain);
          free(lpHeader->lpszAuditEMail);
          free(lpHeader->lpszWebDocsDir);
          free(lpHeader->lpszBaseDir);
          free(lpHeader);
          return -2;
        }
        v2len -= len;
      }
      bEncryptEnable = bEncrypt;
    } else
    {
      /* Allocate empty strings for the web docs and base directories in addition to the audit e-mail */
      lpHeader->lpszWebDocsDir = (LPTSTR) malloc(1);
      lpHeader->lpszBaseDir = (LPTSTR) malloc(1);
      lpHeader->lpszWebDocsDir[0] = 0;
      lpHeader->lpszBaseDir[0] = 0;
    }

    /* If the AuditEMail field was not allocated, then do it now */
    if (!lpHeader->lpszAuditEMail) {
      lpHeader->lpszAuditEMail = (LPTSTR) malloc(1);
      lpHeader->lpszAuditEMail[0] = 0;
    }

    /* If the EMailDomain field was not allocated, then do it now */
    if (!lpHeader->lpszEMailDomain) {
      lpHeader->lpszEMailDomain = (LPTSTR) malloc(1);
      lpHeader->lpszEMailDomain[0] = 0;
    }

    /* If the DefaultURL field was not allocated, then do it now */
    if (!lpHeader->lpszDefaultURL) {
      lpHeader->lpszDefaultURL = (LPTSTR) malloc(1);
      lpHeader->lpszDefaultURL[0] = 0;
    }

    /* Skip ahead for the remainder of the entry, if anything exists */
    if (v2len)
      fseek(index, v2len, SEEK_CUR);

    /* Return zero for the header type */
    return 0;
  }

  /* Check the index entry type */
  if (strcmp(type, "@1") == 0)
  {
    /* This is a version 1 index entry :-)! */
    LPDIRLIST_ENTRY lpDirEntry = (LPDIRLIST_ENTRY) malloc(sizeof(DIRLIST_ENTRY));
    LPTSTR lpszRealName;
    LPTSTR loc;
    LPTSTR lpszUserName;
    LPTSTR lpszLastName;
    DWORD len;
    memset(lpDirEntry, (char) NULL, sizeof(DIRLIST_ENTRY));
    *lpData = (LPBYTE) lpDirEntry;

    /* Read the length, allocate the memory, and read the data */
    if (!ReadDWORDNoEncrypt(index, &len)) {
      free(lpDirEntry);
      return -2;
    }
    lpszRealName = malloc(len + 1);
    if (fread(lpszRealName, len + 1, 1, index) < 1) {
      free(lpszRealName);
      free(lpDirEntry);
      return -2;
    }

    /* Split the real name into the first and last names */
    lpszLastName = GetLastName(lpszRealName);
    if (lpszLastName) {
      lpDirEntry->lpszLastName = (LPTSTR) malloc(strlen(lpszLastName) + 1);
      strcpy(lpDirEntry->lpszLastName, lpszLastName);
      lpDirEntry->lpszFirstName = lpszRealName;
      lpszLastName[0] = 0;
      lpDirEntry->lpszFirstName = (LPTSTR) realloc(lpDirEntry->lpszFirstName, lpszLastName - lpszRealName);
      lpDirEntry->lpszFirstName[lpszLastName - lpszRealName - 1] = 0;
    } else {
      lpDirEntry->lpszLastName = lpszRealName;
      lpDirEntry->lpszFirstName = (LPTSTR) malloc(1);
      lpDirEntry->lpszFirstName[0] = 0;
    }

    /* Read the length, allocate the memory, and read the data */
    if (!ReadDWORDNoEncrypt(index, &len)) {
      free(lpDirEntry->lpszFirstName);
      free(lpDirEntry->lpszLastName);
      free(lpDirEntry);
      return -2;
    }
    lpDirEntry->lpszWebSite = malloc(len + 1);
    if (fread(lpDirEntry->lpszWebSite, len + 1, 1, index) < 1) {
      free(lpDirEntry->lpszFirstName);
      free(lpDirEntry->lpszLastName);
      free(lpDirEntry->lpszWebSite);
      free(lpDirEntry);
      return -2;
    }

    /* Guess the username from the web site address */
    lpszUserName = (LPTSTR) malloc(strlen(lpDirEntry->lpszWebSite) + 1);
    strcpy(lpszUserName, lpDirEntry->lpszWebSite);
    loc = strchr(lpszUserName, '~');
    if (loc) {
      P_TYPE lenCPU;
      loc = (char *) (loc + 1);
      lenCPU = (P_TYPE) strrchr(loc, '/');
      if (lenCPU)
        len = (DWORD) ((lenCPU - (P_TYPE) loc) & 0xFFFFFFFF);
      else
        len = strlen(loc);
      loc[len] = 0;
    } else
      loc = lpszUserName;

    /* Set the structure's username and clean up our temporary allocation */
    lpDirEntry->lpszUserName = (LPTSTR) malloc(strlen(loc) + 1);
    strcpy(lpDirEntry->lpszUserName, loc);
    free(lpszUserName);

    /* Fill in the default values for the other strings */
    lpDirEntry->lpszEMail = malloc(strlen(lpDirEntry->lpszUserName) + strlen(DIRLIST_EMAILDOMAIN) + 2);
    sprintf(lpDirEntry->lpszEMail, "%s@%s", lpDirEntry->lpszUserName, DIRLIST_EMAILDOMAIN);
    lpDirEntry->lpszTelephone = malloc(1);
    lpDirEntry->lpszTelephone[0] = 0;
    lpDirEntry->nClassDataValues = 0;
    lpDirEntry->lpClassData = NULL;

    /* Return 1 for the index entry type */
    return 1;
  }

  if (strcmp(type, "@2") == 0)
  {
    /* Version 2 index entry :-)!  Horray for v2! */
    DWORD v2len;
    LPDIRLIST_ENTRY lpDirEntry = (LPDIRLIST_ENTRY) malloc(sizeof(DIRLIST_ENTRY));
    memset(lpDirEntry, (char) NULL, sizeof(DIRLIST_ENTRY));
    *lpData = (LPBYTE) lpDirEntry;
    if (!ReadDWORD(index, &v2len)) {
      free(lpDirEntry);
      return -2;
    }

    /* Read the flags */
    if (!ReadDWORD(index, &(lpDirEntry->dwFlags))) {
      free(lpDirEntry);
      return -2;
    }
    v2len -= sizeof(DWORD);

    /* Read the strings */
    if (!(len = ReadString(index, &(lpDirEntry->lpszUserName)))) {
      free(lpDirEntry);
      return -2;
    }
    v2len -= len;
    if (!(len = ReadString(index, &(lpDirEntry->lpszFirstName)))) {
      free(lpDirEntry->lpszFirstName);
      free(lpDirEntry);
      return -2;
    }
    v2len -= len;
    if (!(len = ReadString(index, &(lpDirEntry->lpszLastName)))) {
      free(lpDirEntry->lpszLastName);
      free(lpDirEntry);
      return -2;
    }
    v2len -= len;
    if (!(len = ReadString(index, &(lpDirEntry->lpszWebSite)))) {
      free(lpDirEntry->lpszUserName);
      free(lpDirEntry->lpszFirstName);
      free(lpDirEntry->lpszLastName);
      free(lpDirEntry);
      return -2;
    }
    v2len -= len;
    if (!(len = ReadString(index, &(lpDirEntry->lpszEMail)))) {
      free(lpDirEntry->lpszUserName);
      free(lpDirEntry->lpszFirstName);
      free(lpDirEntry->lpszLastName);
      free(lpDirEntry->lpszWebSite);
      free(lpDirEntry);
      return -2;
    }
    v2len -= len;
    if (!(len = ReadString(index, &(lpDirEntry->lpszTelephone)))) {
      free(lpDirEntry->lpszUserName);
      free(lpDirEntry->lpszFirstName);
      free(lpDirEntry->lpszLastName);
      free(lpDirEntry->lpszWebSite);
      free(lpDirEntry->lpszEMail);
      free(lpDirEntry);
      return -2;
    }
    v2len -= len;

    /* If FLAGS_CLASSDATA is set, and if there are bytes left, read the class data */
    if (v2len && (lpDirEntry->dwFlags & FLAGS_CLASSDATA))
    {
      /* Read the number of class data entires */
      if (!ReadDWORD(index, &(lpDirEntry->nClassDataValues))) {
        free(lpDirEntry->lpszUserName);
        free(lpDirEntry->lpszFirstName);
        free(lpDirEntry->lpszLastName);
        free(lpDirEntry->lpszWebSite);
        free(lpDirEntry->lpszEMail);
        free(lpDirEntry->lpszTelephone);
        free(lpDirEntry);
        return -2;
      }
      v2len -= sizeof(DWORD);

      /* If it is above zero, allocate memory for the class data */
      if (lpDirEntry->nClassDataValues > 0) {
        DWORD i;
        lpDirEntry->lpClassData = (LPDIRLIST_CLASSDATA) malloc(sizeof(DIRLIST_CLASSDATA) * lpDirEntry->nClassDataValues);

        /* Read each class data entry */
        for (i = 0; i < lpDirEntry->nClassDataValues; i++)
        {
          /* General data in every class data entry */
          if (!ReadDWORD(index, &(lpDirEntry->lpClassData[i].idClass)))
            return -2;
          v2len -= sizeof(DWORD);
          if (!ReadDWORD(index, &(lpDirEntry->lpClassData[i].dwFlags)))
            return -2;
          v2len -= sizeof(DWORD);

          /* Get the value, depending on whether a string or an ordinal is stored */
          if (lpDirEntry->lpClassData[i].dwFlags & CLASSDATA_OTHER)
          {
            /* String data */
            lpDirEntry->lpClassData[i].dwValue = 0;
            if (!(len = ReadString(index, &(lpDirEntry->lpClassData[i].lpszValue))))
              return -2;
            v2len -= len;
          } else
          {
            /* Ordinal data -- get the value and then attempt to get the string */
            LPDIRLIST_CLASS lpClass;
            if (!ReadDWORD(index, &(lpDirEntry->lpClassData[i].dwValue)))
              return -2;
            v2len -= sizeof(DWORD);

            /* Query the class by its ID */
            lpDirEntry->lpClassData[i].lpszValue = NULL;
            lpClass = GetClass(lpClassList, lpDirEntry->lpClassData[i].idClass);
            if (lpClass)
            {
              /* Query the class value for dwValue */
              LPDIRLIST_CLASSVALUE lpClassValue = GetClassValue(lpClass, lpDirEntry->lpClassData[i].dwValue);
              if (lpClassValue)
                lpDirEntry->lpClassData[i].lpszValue = lpClassValue->lpszValue;
            }
          }
        }
      }
    } else
    {
      /* Nullify the class data information */
      lpDirEntry->nClassDataValues = 0;
      lpDirEntry->lpClassData = NULL;
    }

    /* Skip ahead for the remainder of the entry, if anything exists */
    if (v2len)
      SeekFile(index, v2len, SEEK_CUR);
    return 2;
  }

  if (strcmp(type, "@3") == 0)
  {
    /* Class definition and values index entry */
    DWORD v2len;
    DWORD dwReserved;
    LPDIRLIST_CLASS lpClass = (LPDIRLIST_CLASS) malloc(sizeof(DIRLIST_CLASS));
    memset(lpClass, (char) NULL, sizeof(DIRLIST_CLASS));
    *lpData = (LPBYTE) lpClass;
    if (!ReadDWORD(index, &v2len)) {
      free(lpClass);
      return -2;
    }

    /* Read the class information */
    if (!ReadDWORD(index, &(lpClass->idClass))) {
      free(lpClass);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (!(len = ReadString(index, &(lpClass->lpszName)))) {
      free(lpClass);
      return -2;
    }
    v2len -= len;
    if (!ReadDWORD(index, &(lpClass->nValues))) {
      free(lpClass->lpszName);
      free(lpClass);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (!ReadDWORD(index, &(lpClass->dwFlags))) {
      free(lpClass->lpszName);
      free(lpClass);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (!ReadDWORD(index, &(lpClass->idDefaultValue))) {
      free(lpClass->lpszName);
      free(lpClass);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (!ReadDWORD(index, &(dwReserved))) {
      free(lpClass->lpszName);
      free(lpClass);
      return -2;
    }
    v2len -= sizeof(DWORD);

    /* Read the class values, if any */
    if (lpClass->nValues)
    {
      /* Allocate the memory */
      DWORD i;
      lpClass->lpValues = (LPDIRLIST_CLASSVALUE) malloc(sizeof(DIRLIST_CLASSVALUE) * lpClass->nValues);

      /* Read each class value from the file */
      for (i = 0; i < lpClass->nValues; i++)
      {
        /* Value ID */
        if (!ReadDWORD(index, &(lpClass->lpValues[i].idValue)))
          return -2;
        v2len -= sizeof(DWORD);

        /* Associated string */
        if (!(len = ReadString(index, &(lpClass->lpValues[i].lpszValue))))
          return -2;
        v2len -= len;
      }
    } else
    {
      /* Nullify the value list pointer */
      lpClass->lpValues = NULL;
    }

    /* Skip ahead for the remainder of the entry, if anything exists */
    if (v2len)
      SeekFile(index, v2len, SEEK_CUR);
    return 3;
  }

  if (strcmp(type, "@4") == 0)
  {
    /* Sort table (read the header only) */
    BOOL bEncrypt = bEncryptEnable;
    DWORD v2len;
    LPDIRLIST_SORTTABLE lpTable = (LPDIRLIST_SORTTABLE) malloc(sizeof(DIRLIST_SORTTABLE));
    memset(lpTable, (char) NULL, sizeof(DIRLIST_SORTTABLE));
    *lpData = (LPBYTE) lpTable;
    if (!ReadDWORDNoEncrypt(index, &v2len)) {
      free(lpTable);
      return -2;
    }

    /* Read the sort table information */
    if (!ReadDWORDNoEncrypt(index, &(lpTable->idSort))) {
      free(lpTable);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (!ReadDWORDNoEncrypt(index, &(lpTable->nEntries))) {
      free(lpTable);
      return -2;
    }
    v2len -= sizeof(DWORD);
    bEncryptEnable = FALSE;
    if (!(len = ReadString(index, &(lpTable->lpszOrder)))) {
      free(lpTable);
      return -2;
    }
    v2len -= len;
    bEncryptEnable = bEncrypt;

    /* Set the file position to the start of the sort table entries */
    lpTable->dwFilePos = (DWORD) ftell(index);

    /* Skip ahead for the remainder of the entry */
    if (v2len)
      fseek(index, v2len, SEEK_CUR);
    return 4;
  }

  if (strcmp(type, "@5") == 0)
  {
    /* User information */
    DWORD v2len;
    LPDIRLIST_USER lpUser = (LPDIRLIST_USER) malloc(sizeof(DIRLIST_USER));
    memset(lpUser, (char) NULL, sizeof(DIRLIST_USER));
    *lpData = (LPBYTE) lpUser;
    if (!ReadDWORD(index, &v2len)) {
      free(lpUser);
      return -2;
    }

    /* Read the user information */
    if (!ReadDWORD(index, &(lpUser->idUser))) {
      free(lpUser);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (!(len = ReadString(index, &(lpUser->lpszName)))) {
      free(lpUser);
      return -2;
    }
    v2len -= len;
    if (!(len = ReadString(index, &(lpUser->lpszPassword)))) {
      free(lpUser->lpszName);
      free(lpUser);
      return -2;
    }
    v2len -= len;
    if (!(len = ReadString(index, &(lpUser->lpszDescription)))) {
      free(lpUser->lpszName);
      free(lpUser->lpszPassword);
      free(lpUser);
      return -2;
    }
    v2len -= len;

    /* Read the user level and permitted operations */
    if (!ReadDWORD(index, &(lpUser->dwUserLevel))) {
      free(lpUser->lpszName);
      free(lpUser->lpszPassword);
      free(lpUser->lpszDescription);
      free(lpUser);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (!ReadDWORD(index, &(lpUser->dwOperations))) {
      free(lpUser->lpszName);
      free(lpUser->lpszPassword);
      free(lpUser->lpszDescription);
      free(lpUser);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (ReadFile(&(lpUser->bAdminMode), sizeof(BOOL), 1, index) < 1) {
      free(lpUser->lpszName);
      free(lpUser->lpszPassword);
      free(lpUser->lpszDescription);
      free(lpUser);
      return -2;
    }
    v2len -= sizeof(BOOL);
    if (!ReadDWORD(index, &(lpUser->dwSearchLevel))) {
      free(lpUser->lpszName);
      free(lpUser->lpszPassword);
      free(lpUser->lpszDescription);
      free(lpUser);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (ReadFile(&(lpUser->bAudit), sizeof(BOOL), 1, index) < 1) {
      free(lpUser->lpszName);
      free(lpUser->lpszPassword);
      free(lpUser->lpszDescription);
      free(lpUser);
      return -2;
    }
    v2len -= sizeof(BOOL);
    if (!ReadDWORD(index, &(lpUser->dwAuditOperations))) {
      free(lpUser->lpszName);
      free(lpUser->lpszPassword);
      free(lpUser->lpszDescription);
      free(lpUser);
      return -2;
    }
    v2len -= sizeof(DWORD);

    /* Skip ahead for the remainder of the entry */
    if (v2len)
      SeekFile(index, v2len, SEEK_CUR);
    return 5;
  }

  if (strcmp(type, "@6") == 0)
  {
    /* Locking information record (only one per file is permitted in this version) */
    DWORD v2len;
    LPDIRLIST_LOCK lpLock = (LPDIRLIST_LOCK) malloc(sizeof(DIRLIST_LOCK));
    memset(lpLock, (char) NULL, sizeof(DIRLIST_LOCK));
    *lpData = (LPBYTE) lpLock;
    if (!ReadDWORD(index, &v2len)) {
      free(lpLock);
      return -2;
    }

    /* Read the locking information */
    if (!ReadDWORD(index, &(lpLock->dwFilePosUser))) {
      free(lpLock);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (!ReadDWORD(index, &(lpLock->dwOperations))) {
      free(lpLock);
      return -2;
    }
    v2len -= sizeof(DWORD);
    if (ReadFile(&(lpLock->bAdminMode), sizeof(BOOL), 1, index) < 1) {
      free(lpLock);
      return -2;
    }
    v2len -= sizeof(BOOL);

    /* If the lock time structure is larger than the rest of the bytes (likely an import),
       then assume a time of zero */
    if (sizeof(time_t) + sizeof(DWORD) > v2len) {
      lpLock->tLockTime = 0;
      SeekFile(index, v2len - sizeof(DWORD), SEEK_CUR);
    } else
    {
      /* Read the lock time */
      if (ReadFile(&(lpLock->tLockTime), sizeof(time_t), 1, index) < 1) {
        free(lpLock);
        return -2;
      }

      /* If we are importing from a different endian, don't bother swabbing */
      if (bConvertEndian)
        lpLock->tLockTime = 0;
      v2len -= sizeof(time_t);
    }
    if (!ReadDWORD(index, &(lpLock->dwLockDuration))) {
      free(lpLock);
      return -2;
    }
    v2len -= sizeof(DWORD);

    /* Skip ahead for the remainder of the entry */
    if (v2len)
      SeekFile(index, v2len, SEEK_CUR);
    return 6;
  }

  /* Unrecognized index entry type!  Read the length and skip it */
  if (!ReadDWORD(index, &len))
    return -2;
  SeekFile(index, len, SEEK_CUR);
  return -1;
}

void WriteIndexHeader(FILE *index)
{
  /* Write the file header */
  char header[25];
  strcpy(header, "#!/dirlist");
  strcpy((LPTSTR)(header + 11), "--!header_002");
  fwrite(header, 1, 25, index);
}

DWORD CheckIndexHeader(FILE *index)
{
  /* Check the file header */
  char test1[25], test2[25];
  fread(test1, 1, 25, index);
  strcpy(test2, "#!/dirlist");
  strcpy((LPTSTR)(test2 + 11), "--!header_");
  if (memcmp(test1, test2, 21) == 0)
  {
    /* Return the version */
    return strtoul((LPTSTR) (test1 + 21), NULL, 10);
  }

  /* Unsupported file type, or version is too high */
  return FALSE;
}

BOOL CheckUserDirectory(LPTSTR lpszCheckFile, BOOL bVerifyHTML, LPCTSTR lpszFormatStr, ...)
{
  /* Get the ending parameters (used when we write the inside data to the string) */
  FILE *testFile;
  va_list vaList;
  va_start(vaList, lpszFormatStr);

  /* Write the formatted arguments to the lpszCheckFile string (caller assured that lpszCheckFile met the required length to avoid buffer overrun) */
  vsprintf(lpszCheckFile, lpszFormatStr, vaList);
  va_end(vaList);

  /* Try to open this file */
  if ((testFile = fopen(lpszCheckFile, "r")) > 0)
  {
    /* Check the HTML for validity and return it's result (file is closed by the function) */
    if (!bVerifyHTML) {
      fclose(testFile);
      return TRUE;
    } else
      return VerifyHTML(testFile);
  }

  /* File doesn't exist..  Return -1 */
  return -1;
}

/* Reserve space in the index file with null data */
BOOL ReserveSpace(FILE *index, DWORD nBytes)
{
  /* Compute length */
  DWORD len = nBytes - (sizeof(DWORD) + 2);
  LPBYTE lpNull[1024];

  /* Write entry header of type R */
  fwrite("@R", 1, 2, index);

  /* Write the total length of the entry */
  fwrite(&len, sizeof(DWORD), 1, index);

  /* Write the blank space */
  memset(lpNull, 0, 1024);
  while (len) {
    if (len / 1024 > 0) {
      fwrite(lpNull, 1024, 1, index);
      len -= 1024;
    } else {
      fwrite(lpNull, len, 1, index);
      len = 0;
    }
  }

  /* We did it :-) */
  return TRUE;
}

/* Load the index file into memory */
BOOL LoadIndexFile(LPCTSTR lpszIndexName, LPDIRLIST_INDEX lpIndex, BOOL bManualCount, BOOL bEmitStatus)
{
  /* Load the directory into memory while sorting it */
  LPDIRLIST_HEADER *lpHeader;
  LPDIRLIST_ENTRY *lpDirEntry;
  LPDIRLIST_CLASS *lpClass;
  LPDIRLIST_SORTTABLE *lpTable;
  LPDIRLIST_USER *lpUser;
  LPDIRLIST_LOCK *lpLock;
  FILE *index;
  int hIndex;

  /* If lpIndex is NULL, then return FALSE */
  if (!lpIndex) {
    if (bEmitStatus)
      infprintf("#Could not load the index because NULL was passed as the pointer to place the data into!\n");
    return FALSE;
  }

  /* Link the pointers to the destinations (temporary fixup for the lpIndex conversion) */
  lpHeader = &(lpIndex->lpHeader);
  lpDirEntry = &(lpIndex->lpDirEntry);
  lpClass = &(lpIndex->lpClass);
  lpTable = &(lpIndex->lpSortTable);
  lpUser = &(lpIndex->lpUser);
  lpLock = &(lpIndex->lpLock);

  /* Load the source file into memory, if any */
  hIndex = open(lpszIndexName, O_RDONLY, S_IREAD);
  if (hIndex >= 0)
  {
    /* Check the file header */
    DWORD ver;
    index = fdopen(hIndex, "r");
    if (!(ver = CheckIndexHeader(index))) {
      if (bEmitStatus)
        infprintf("#Index file version unknown or invalid.\n");
      return FALSE;
    }
    else
    {
      /* Update the user on our status */
      LPDIRLIST_ENTRY lpPrevDirEntry = NULL;
      LPDIRLIST_ENTRY lpCurrent = NULL;
      LPDIRLIST_CLASS lpPrevClass = NULL;
      LPDIRLIST_SORTTABLE lpPrevTable = NULL;
      LPDIRLIST_USER lpPrevUser = NULL;
      LPDIRLIST_HEADER lpBackupHeader = NULL;
      LPDIRLIST_CLASS lpBackupClass = NULL;
      BOOL verOk = TRUE;
      LPBYTE lpData;
      int type;

      /* If the header given was NULL, then use our backup header */
      if (!lpHeader) {
        lpHeader = &lpBackupHeader;
        *lpHeader = NULL;
      }

      /* If the header already has an allocation, then free it! */
      if (*lpHeader) {
        FreeHeader(*lpHeader);
        free(*lpHeader);
        *lpHeader = NULL;
      }

      /* If this is a version 2 header, then read the header entry */
      if (ver > 2)
        verOk = FALSE;
      if (ver == 2) {
        if (ReadIndexEntry(index, &lpData, NULL) == 0)
        {
          /* Check the version information, ensuring we were meant to read this file */
          *lpHeader = (LPDIRLIST_HEADER) lpData;
          if ((*lpHeader)->dwLowestVersion > DIRLIST_VERSION_MINOR)
            verOk = FALSE;

          /* If encryption is set, then enable encryption */
          if ((*lpHeader)->dwFlags & HEADER_ENCRYPTFILE)
            bEncryptEnable = TRUE;
          else
            bEncryptEnable = FALSE;
        } else
        {
          /* Not a valid version 2 index file (the @0 is _required_)! */
          printf("#Invalid version 2 index file!\n");
          verOk = FALSE;
        }
      }

      /* Initialize the counters to NULL if we are not manually counting */
      if (!bManualCount) {
        lpIndex->lpnDirEntries = NULL;
        lpIndex->lpnClasses = NULL;
        lpIndex->lpnTables = NULL;
        lpIndex->lpnUsers = NULL;
        lpIndex->lpnLocks = NULL;
      }

      /* Make sure we should run this section of code */
      if (verOk && (lpDirEntry || lpClass || lpTable || lpUser || lpLock || ((bManualCount)? lpIndex->lpnDirEntries || lpIndex->lpnClasses || lpIndex->lpnTables || lpIndex->lpnUsers || lpIndex->lpnLocks: TRUE))) {
        if (bEmitStatus)
          infprintf("#Reading index file into memory (file version=%u)\n", ver);

        /* If the header was allocated using our own allocation, then free it here */
        if (lpBackupHeader) {
          FreeHeader(*lpHeader);
          free(*lpHeader);
        }

        /* If the class was not given, then have it point to the backup */
        if (!lpClass) {
          lpClass = &lpBackupClass;
          *lpClass = NULL;
        }

        /* Loop through the file */
        while ((type = ReadIndexEntry(index, &lpData, *lpClass)) >= -1) {
          switch (type) {

          case 1:
          case 2:

            /* Directory list entry -- add it to the list :-)! */
            if (lpDirEntry)
            {
              if (FetchEntry(lpDirEntry, &lpPrevDirEntry, &lpCurrent, ((LPDIRLIST_ENTRY) lpData)->lpszUserName, FALSE))
              {
                /* Remove the current entry and replace it with ours :-) */
                FreeEntry(lpCurrent);
                if (lpPrevDirEntry) {
                  lpPrevDirEntry->next = (LPBYTE) lpData;
                  ((LPDIRLIST_ENTRY) lpData)->next = lpCurrent->next;
                } else
                {
                  *lpDirEntry = (LPDIRLIST_ENTRY) lpData;
                  (*lpDirEntry)->next = lpCurrent->next;
                }
                free(lpCurrent);
                lpCurrent = (LPDIRLIST_ENTRY) lpData;
              } else {
                if (lpPrevDirEntry) {
                  lpPrevDirEntry->next = (LPBYTE) lpData;
                  if (lpCurrent)
                    ((LPDIRLIST_ENTRY) lpData)->next = (LPBYTE) lpCurrent;
                  else
                    ((LPDIRLIST_ENTRY) lpData)->next = NULL;
                  lpCurrent = (LPDIRLIST_ENTRY) lpData;
                } else
                {
                  *lpDirEntry = (LPDIRLIST_ENTRY) lpData;
                  if (lpCurrent)
                    (*lpDirEntry)->next = (LPBYTE) lpCurrent;
                  else
                    (*lpDirEntry)->next = NULL;
                  lpCurrent = *lpDirEntry;
                }
              }
            } else
            {
              /* Free all entries; we don't care about them */
              FreeEntry((LPDIRLIST_ENTRY) lpData);
              free(lpData);
            }

            /* Increment the counter */
            if (lpIndex->lpnDirEntries)
              (*(lpIndex->lpnDirEntries))++;
            break;

          case 3:

            /* Class information entry -- add it to the list :-)! */
            if (lpClass)
            {
              ((LPDIRLIST_CLASS) lpData)->next = NULL;
              if (lpPrevClass)
                lpPrevClass->next = (LPBYTE) lpData;
              else
                *lpClass = (LPDIRLIST_CLASS) lpData;
              lpPrevClass = (LPDIRLIST_CLASS) lpData;
              lpPrevClass->next = NULL;
            } else
            {
              /* Free the class; we don't care about it */
              ((LPDIRLIST_CLASS) lpData)->next = NULL;
              FreeClasses((LPDIRLIST_CLASS) lpData);
            }

            /* Increment the counter */
            if (lpIndex->lpnClasses)
              (*(lpIndex->lpnClasses))++;
            break;

          case 4:

            /* Sort table */
            if (lpTable)
            {
              ((LPDIRLIST_SORTTABLE) lpData)->next = NULL;
              if (lpPrevTable)
                lpPrevTable->next = (LPBYTE) lpData;
              else
                *lpTable = (LPDIRLIST_SORTTABLE) lpData;
              lpPrevTable = (LPDIRLIST_SORTTABLE) lpData;
              lpPrevTable->next = NULL;
            } else
            {
              /* Free the sort table; we don't care about it */
              ((LPDIRLIST_SORTTABLE) lpData)->next = NULL;
              FreeSortTables((LPDIRLIST_SORTTABLE) lpData);
            }

            /* Increment the counter */
            if (lpIndex->lpnTables)
              (*(lpIndex->lpnTables))++;
            break;

          case 5:

            /* User information */
            if (lpUser)
            {
              ((LPDIRLIST_USER) lpData)->next = NULL;
              if (lpPrevUser)
                lpPrevUser->next = (LPBYTE) lpData;
              else
                *lpUser = (LPDIRLIST_USER) lpData;
              lpPrevUser = (LPDIRLIST_USER) lpData;
              lpPrevUser->next = NULL;
            } else
            {
              /* Free the user; we're not collecting user info */
              ((LPDIRLIST_USER) lpData)->next = NULL;
              FreeUsers((LPDIRLIST_USER) lpData);
            }

            /* Increment the counter */
            if (lpIndex->lpnUsers)
              (*(lpIndex->lpnUsers))++;
            break;

          case 6:

            /* Index locking information (only one can exist at a time; the last one is used) */
            if (lpLock)
              free(lpLock);
            lpIndex->lpLock = (LPDIRLIST_LOCK) lpData;
            *(lpIndex->lpnLocks) = 1;
          }
          fflush(stdout);
        }

        /* If we had a read error, return false */
        if (type == -2) {
          if (bEmitStatus)
            infprintf("#Read error during entry fetch\n");
          fclose(index);
          FreeIndex(lpIndex);
          return FALSE;
        }

        /* Link the counter variables to the header if we did not count manually */
        if (!bManualCount) {
          lpIndex->lpnDirEntries = &((*lpHeader)->nType2Entries);
          lpIndex->lpnClasses = &((*lpHeader)->nType3Entries);
          lpIndex->lpnTables = &((*lpHeader)->nType4Entries);
          lpIndex->lpnUsers = &((*lpHeader)->nType5Entries);
          lpIndex->lpnLocks = &((*lpHeader)->nType6Entries);
        }
      }
      if (!verOk)
      {
        /* Index file cannot be read by us due to versioning -- tell the user */
        if (bEmitStatus && (*lpHeader))
          infprintf("#Program version not high enough -- upgrade to read this file (File Major: %u, File Minor: %u, File Minimum Minor: %u, Program Major: %u, Program Minor %u)\n", ver, (*lpHeader)->dwVersion, (*lpHeader)->dwLowestVersion, DIRLIST_VERSION_MAJOR, DIRLIST_VERSION_MINOR);
        if (bEmitStatus && !(*lpHeader))
          infprintf("#Program version not high enough -- upgrade to read this file (File Major: %u, Program Major: %u, Program Minor %u)\n", ver, DIRLIST_VERSION_MAJOR, DIRLIST_VERSION_MINOR);
        fclose(index);
        return FALSE;
      }
    }

    /* Close the index */
    fclose(index);
  } else {
    if (bEmitStatus) {
      DWORD origSafeStringsLen = safeStringsLen;
      infprintf("#Could not obtain read access to the source index \"%s\" (it might not exist).\n",
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpszIndexName, STR_NUL_TERM));
      safeStringsLen = origSafeStringsLen;
    }
    return FALSE;
  }

  /* File is now completely in memory (or parts if some params were NULL :-)! */
  return TRUE;
}

/* Save the memory linked lists into an idx file */
BOOL SaveIndexFile(LPCTSTR lpszIndexName, LPDIRLIST_INDEX lpIndex, BOOL bEmitStatus, BOOL bVeryVerbose)
{
  /* Open the index file for output */
  FILE *index;
  int hIndex;
  hIndex = open(lpszIndexName, O_RDWR | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE);
  index = fdopen(hIndex, "w+");
  if (index)
  {
    /* Write the file header and compute the @0 entry */
    LPDIRLIST_ENTRY lpCurDirEntry = lpIndex->lpDirEntry;
    LPDIRLIST_CLASS lpCurClass = lpIndex->lpClass;
    LPDIRLIST_SORTTABLE lpCurTable = lpIndex->lpSortTable;
    LPDIRLIST_USER lpCurUser = lpIndex->lpUser;
    BOOL bFreeHeader = FALSE;
    DWORD dwSortTablePos, dwHeaderPos, nBytes;
    DWORD nDirEntries, nClasses, nTables, nUsers, nLocks;

    /* If we weren't given a header allocation, then make one that we'll free later */
    if (!lpIndex->lpHeader) {
      lpIndex->lpHeader = (LPDIRLIST_HEADER) malloc(sizeof(DIRLIST_HEADER));
      lpIndex->lpHeader->lpszWebDocsDir = (LPTSTR) strdup(DIRLIST_WEBDOCS);
      lpIndex->lpHeader->lpszBaseDir = (LPTSTR) malloc(1);
      lpIndex->lpHeader->lpszAuditEMail = (LPTSTR) malloc(1);
      lpIndex->lpHeader->lpszEMailDomain = (LPTSTR) strdup(DIRLIST_EMAILDOMAIN);
      lpIndex->lpHeader->lpszDefaultURL = (LPTSTR) malloc(1);
      lpIndex->lpHeader->lpszBaseDir[0] = 0;
      lpIndex->lpHeader->lpszAuditEMail[0] = 0;
      lpIndex->lpHeader->lpszDefaultURL[0] = 0;
      bFreeHeader = TRUE;
    }

    /* Ensure sequence numbers and locking are enabled */
    lpIndex->lpHeader->dwFlags |= HEADER_SEQUENCENUMBERS | HEADER_ENFORCELOCKS;

    /* If encryption is enabled, then set it */
    if (lpIndex->lpHeader->dwFlags & HEADER_ENCRYPTFILE) {
#ifdef DIRLIST_USE_PACKETMANGLER
      bEncryptEnable = TRUE;
      wEncryptMethod = ENCRYPT_PACKETMANGLE_A;
#else
      bEncryptEnable = FALSE;
      wEncryptMethod = ENCRYPT_NONE;
#endif
    } else {
      bEncryptEnable = FALSE;
      wEncryptMethod = ENCRYPT_NONE;
    }

    /* Set the header data */
    lpIndex->lpHeader->dwVersion = DIRLIST_VERSION_MINOR;
    lpIndex->lpHeader->dwLowestVersion = DIRLIST_VERSION_FILELOWEST;
    lpIndex->lpHeader->nEntryTypes = 6;
    lpIndex->lpHeader->nType1Entries = 0;
    lpIndex->lpHeader->nType2Entries = 0;
    lpIndex->lpHeader->nType3Entries = 0;
    lpIndex->lpHeader->nType4Entries = 0;
    lpIndex->lpHeader->nType5Entries = 0;
    lpIndex->lpHeader->nType6Entries = 0;

    /* Display our status */
    if (bEmitStatus) {
      infprintf("#Output index file opened; Writing header...\n");
      fflush(stdout);
    }

    /* Count the index entries */
    nDirEntries = 0;
    while (lpCurDirEntry) {
      nDirEntries++;
      lpCurDirEntry = (LPDIRLIST_ENTRY) lpCurDirEntry->next;
    }

    /* Count the class definitions */
    nClasses = 0;
    while (lpCurClass) {
      nClasses++;
      lpCurClass = (LPDIRLIST_CLASS) lpCurClass->next;
    }

    /* Count the sort tables */
    nTables = 0;
    while (lpCurTable) {
      nTables++;
      lpCurTable = (LPDIRLIST_SORTTABLE) lpCurTable->next;
    }

    /* Count the users */
    nUsers = 0;
    while (lpCurUser) {
      nUsers++;
      lpCurUser = (LPDIRLIST_USER) lpCurUser->next;
    }

    /* "Count" the locks (only one or zero can exist in DirList2) */
    if (lpIndex->lpLock)
      nLocks = 1;
    else
      nLocks = 0;

    /* Increase the sequence number */
    lpIndex->lpHeader->dwSequenceNumber++;

    /* Write the complete header */
    lpIndex->lpHeader->nType2Entries = nDirEntries;
    lpIndex->lpHeader->nType3Entries = nClasses;
    lpIndex->lpHeader->nType4Entries = nTables;
    lpIndex->lpHeader->nType5Entries = nUsers;
    WriteIndexHeader(index);
    dwHeaderPos = ftell(index);
    WriteIndexEntry(index, lpIndex->lpHeader, 0, bVeryVerbose);

    /* Write the class definitions */
    if (bEmitStatus) {
      infprintf("#Writing class definitions\n");
      fflush(stdout);
    }
    lpCurClass = lpIndex->lpClass;
    while (lpCurClass) {
      WriteIndexEntry(index, lpCurClass, 3, bVeryVerbose);
      lpCurClass = (LPDIRLIST_CLASS) lpCurClass->next;
    }

    /* Write the users */
    if (bEmitStatus) {
      infprintf("#Writing user information\n");
      fflush(stdout);
    }
    lpCurUser = lpIndex->lpUser;
    while (lpCurUser) {
      lpCurUser->dwFilePos = ftell(index);
      WriteIndexEntry(index, lpCurUser, 5, bVeryVerbose);
      lpCurUser = (LPDIRLIST_USER) lpCurUser->next;
    }

    /* Reserve the sort table data space */
    nBytes = 0;
    lpCurTable = lpIndex->lpSortTable;
    while (lpCurTable) {
      nBytes += lpCurTable->nEntries * sizeof(DWORD) + 4 * sizeof(DWORD) + strlen(lpCurTable->lpszOrder) + 3;
      lpCurTable = (LPDIRLIST_SORTTABLE) lpCurTable->next;
    }
    dwSortTablePos = (DWORD) ftell(index);
    if (nBytes) {
      ReserveSpace(index, nBytes);
    }

    /* Write the index entries */
    if (bEmitStatus) {
      infprintf("#Writing index entries...\n");
      fflush(stdout);
    }
    lpCurDirEntry = lpIndex->lpDirEntry;
    while (lpCurDirEntry) {
      lpCurDirEntry->dwFilePos = ftell(index);
      WriteIndexEntry(index, lpCurDirEntry, 2, bVeryVerbose);
      lpCurDirEntry = (LPDIRLIST_ENTRY) lpCurDirEntry->next;
    }

    /* Write the locking information */
    if (bEmitStatus) {
      infprintf("#Writing lock information...\n");
      fflush(stdout);
    }
    if (lpIndex->lpLock)
    {
      /* Write only if a user exists for the lock */
      if (lpIndex->lpLock->lpUser) {
        lpIndex->lpLock->dwFilePosUser = lpIndex->lpLock->lpUser->dwFilePos;
        lpIndex->lpHeader->dwFilePosLock = ftell(index);
        WriteIndexEntry(index, lpIndex->lpLock, 6, bVeryVerbose);

        /* Write the header again, this time with the lock structure position */
        SeekFile(index, (long) dwHeaderPos, SEEK_SET);
        WriteIndexEntry(index, lpIndex->lpHeader, 0, bVeryVerbose);
      }
    }

    /* Write the sort tables */
    SeekFile(index, (long) dwSortTablePos, SEEK_SET);
    if (bEmitStatus) {
      infprintf("#Writing sort tables...\n");
      fflush(stdout);
    }
    lpCurTable = lpIndex->lpSortTable;
    while (lpCurTable) {
      WriteIndexEntry(index, lpCurTable, 4, bVeryVerbose);
      lpCurTable = (LPDIRLIST_SORTTABLE) lpCurTable->next;
    }

    /* Close the file */
    fclose(index);

    /* Free the header if we're supposed to */
    if (bFreeHeader) {
      FreeHeader(lpIndex->lpHeader);
      free(lpIndex->lpHeader);
      lpIndex->lpHeader = NULL;
    }

    /* Display our status */
    if (bEmitStatus) {
      infprintf("#File finalized and closed.\n");
      fflush(stdout);
    }

    /* Return TRUE for success */
    return TRUE;
  } else
  {
    /* Inform the user of this error (file could not be opened) */
    if (bEmitStatus) {
      DWORD origSafeStringsLen = safeStringsLen;
      infprintf("#Output file \"%s\" could not be created/opened for writing!\n",
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpszIndexName, STR_NUL_TERM));
      safeStringsLen = origSafeStringsLen;
    }

    /* Return FALSE for file creation error */
    return FALSE;
  }
}

BOOL ProcessDirectory(LPCTSTR lpszDir, LPDIRLIST_INDEX lpIndex, BOOL bEmitStatus)
{
  /* Load the directory into memory and sort it */
  LPDIRLIST_ENTRY lpCurDirEntry;
  DWORD length;

  if (bEmitStatus) {
    infprintf("#Directory Indexing in Progress:\n");
  }

  /* Set FLAGS_NOHOMEDIR on all entries, which will be unset if a directory exists */
  lpCurDirEntry = lpIndex->lpDirEntry;
  while (lpCurDirEntry) {
    lpCurDirEntry->dwFlags |= FLAGS_NOHOMEDIR;
    lpCurDirEntry = (LPDIRLIST_ENTRY) lpCurDirEntry->next;
  }

  /* Load and merge the directory */
  if (bEmitStatus) {
    DWORD origSafeStringsLen = safeStringsLen;
    infprintf("#Loading \"%s\" and merging it into the current directory list...\n",
              SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpszDir, STR_NUL_TERM));
    safeStringsLen = origSafeStringsLen;
    fflush(stdout);
  }
  length = LoadDirectory(lpszDir, &(lpIndex->lpDirEntry), lpIndex->lpHeader->lpszEMailDomain);

  /* If we have something to do, then continue */
  if (length || *(lpIndex->lpnDirEntries))
  {
    /* Go through each user's directory, looking for a web_docs directory */
    LPDIRLIST_SORTTABLE lpCurTable = lpIndex->lpSortTable;
    DWORD webDocsBaseLength = strlen(lpszDir) + strlen(lpIndex->lpHeader->lpszWebDocsDir) + 4;  /* Three chars plus term. null */
    DWORD webDocsUserLength = 64; /* 64 for username and filename (not combined) */
    DWORD webDocsFileExtLength = 15; /* 15 for file extension (only 6 really needed, but 9 extra doesn't hurt) */
    LPTSTR webDocs = (LPTSTR) malloc(webDocsBaseLength + webDocsUserLength * 2 + webDocsFileExtLength);
    BOOL found;

    /* Perform pre-processing on all the directory entries */
    lpCurDirEntry = lpIndex->lpDirEntry;
    while (lpCurDirEntry)
    {
      /* If web site checking is not disabled, then set the NOWEBSITE flag */
      /* (if a web site exists, this flag is unset later in this function) */
      if (!(lpCurDirEntry->dwFlags & FLAGS_NOWEBCHECK))
        lpCurDirEntry->dwFlags |= FLAGS_NOWEBSITE;
      lpCurDirEntry = (LPDIRLIST_ENTRY) lpCurDirEntry->next;
    }

    if (bEmitStatus)
    {
      DWORD origSafeStringsLen = safeStringsLen;
      infprintf("#Merged %u directory entries with the %u entries in the current index\n", length, *(lpIndex->lpnDirEntries));
      infprintf("#Processing users for the presence of \"%s\" directories (please wait)...\n",
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpIndex->lpHeader->lpszWebDocsDir, STR_NUL_TERM));
      safeStringsLen = origSafeStringsLen;
      fflush(stdout);
    }

    /* Process each user */
    lpCurDirEntry = lpIndex->lpDirEntry;
    while (lpCurDirEntry)
    {
      /* If we can check the web directory of this user, then do so */
      if (!(lpCurDirEntry->dwFlags & FLAGS_NOWEBCHECK))
      {
        /* Increase the size of the string if the username is larger than the current allocation */
        if (strlen(lpCurDirEntry->lpszUserName) > webDocsUserLength) {
          webDocsUserLength = strlen(lpCurDirEntry->lpszUserName) + 64; /* Add 64 characters to the username to reduce potential reallocations */
          webDocs = (LPTSTR) realloc(webDocs, webDocsBaseLength + webDocsUserLength * 2 + webDocsFileExtLength);
        }

        /* Check for various file types until found */
        /* Order: index.shtml index.html, index.php, index.cgi, index.htm, default.htm, username.html */
        found = CheckUserDirectory(webDocs, TRUE, "%s%c%s%c%s%cindex.shtml", lpszDir, cDirDelimiter, lpCurDirEntry->lpszUserName, cDirDelimiter, lpIndex->lpHeader->lpszWebDocsDir, cDirDelimiter);
        if (found > 0)
          sprintf(webDocs, "/~%s", lpCurDirEntry->lpszUserName);
        else {
          BOOL isIndexSHTML = (found < 0)? FALSE: TRUE;
          found = CheckUserDirectory(webDocs, TRUE, "%s%c%s%c%s%cindex.html", lpszDir, cDirDelimiter, lpCurDirEntry->lpszUserName, cDirDelimiter, lpIndex->lpHeader->lpszWebDocsDir, cDirDelimiter);
          if (found > 0) {
            if (isIndexSHTML)
              sprintf(webDocs, "/~%s/index.html", lpCurDirEntry->lpszUserName);
            else
              sprintf(webDocs, "/~%s", lpCurDirEntry->lpszUserName);
          } else {
            BOOL isIndexHTML = (found < 0)? FALSE: TRUE;
            found = CheckUserDirectory(webDocs, FALSE, "%s%c%s%c%s%cindex.php", lpszDir, cDirDelimiter, lpCurDirEntry->lpszUserName, cDirDelimiter, lpIndex->lpHeader->lpszWebDocsDir, cDirDelimiter);
            if (found >= 0) {
              if (isIndexHTML || isIndexSHTML)
                sprintf(webDocs, "/~%s/index.php", lpCurDirEntry->lpszUserName);
              else
                sprintf(webDocs, "/~%s", lpCurDirEntry->lpszUserName);
            } else {
              found = CheckUserDirectory(webDocs, FALSE, "%s%c%s%c%s%cindex.cgi", lpszDir, cDirDelimiter, lpCurDirEntry->lpszUserName, cDirDelimiter, lpIndex->lpHeader->lpszWebDocsDir, cDirDelimiter);
              if (found >= 0) {
                if (isIndexHTML || isIndexSHTML)
                  sprintf(webDocs, "/~%s/index.cgi", lpCurDirEntry->lpszUserName);
                else
                  sprintf(webDocs, "/~%s", lpCurDirEntry->lpszUserName);
              } else {
                found = CheckUserDirectory(webDocs, TRUE, "%s%c%s%c%s%cindex.htm", lpszDir, cDirDelimiter, lpCurDirEntry->lpszUserName, cDirDelimiter, lpIndex->lpHeader->lpszWebDocsDir, cDirDelimiter);
                if (found > 0) {
                  if (isIndexHTML || isIndexSHTML)
                    sprintf(webDocs, "/~%s/index.htm", lpCurDirEntry->lpszUserName);
                  else
                    sprintf(webDocs, "/~%s", lpCurDirEntry->lpszUserName);
                } else {
                  found = CheckUserDirectory(webDocs, TRUE, "%s%c%s%c%s%cdefault.htm", lpszDir, cDirDelimiter, lpCurDirEntry->lpszUserName, cDirDelimiter, lpIndex->lpHeader->lpszWebDocsDir, cDirDelimiter);
                  if (found > 0)
                    sprintf(webDocs, "/~%s/default.htm", lpCurDirEntry->lpszUserName);
                  else {
                    found = CheckUserDirectory(webDocs, TRUE, "%s%c%s%c%s%c%s.html", lpszDir, cDirDelimiter, lpCurDirEntry->lpszUserName, cDirDelimiter, lpIndex->lpHeader->lpszWebDocsDir, cDirDelimiter, lpCurDirEntry->lpszUserName);
                    if (found > 0)
                      sprintf(webDocs, "/~%s/%s.html", lpCurDirEntry->lpszUserName, lpCurDirEntry->lpszUserName);
                  }
                }
              }
            }
          }
        }

        /* If we found it, then set the lpszWebSite variable and indicate that we have a web site */
        if (found == 1) {
          if (lpCurDirEntry->lpszWebSite)
            free(lpCurDirEntry->lpszWebSite);
          lpCurDirEntry->lpszWebSite = (LPTSTR) malloc(strlen(webDocs) + 1);
          strcpy(lpCurDirEntry->lpszWebSite, webDocs);
          lpCurDirEntry->dwFlags &= ~FLAGS_NOWEBSITE;
        }
      }

      /* Next directory entry */
      lpCurDirEntry = (LPDIRLIST_ENTRY) lpCurDirEntry->next;
    }
    free(webDocs);

    /* Refresh the sort tables */
    if (bEmitStatus) {
      infprintf("#Refreshing sort tables...\n");
      fflush(stdout);
    }
    while (lpCurTable) {
      CreateSortTable(lpCurTable, lpIndex->lpDirEntry, lpIndex->lpClass);
      lpCurTable = (LPDIRLIST_SORTTABLE) lpCurTable->next;
    }

    /* We did it :-) */
    return TRUE;
  } else
  {
    /* Tell user we have nothing to do! */
    if (bEmitStatus)
      infprintf("#Nothing to do!  No names exist in the source index (if present), and no directories exist in the given directory system.\n");
  }

  /* Return negative one for not doing anything */
  return -1;
}

/* Read a line from the input stream into lpszLine, changing lpszLine's allocation as needed */
BOOL ReadTextLine(FILE *pInput, LPTSTR *lpszLine, LPDWORD lpdwLineLen, LPDWORD lpdwLineAlloc, LPDWORD lpdwMaxRead)
{
  /* Loop until a newline character is found */
  BOOL bExit = FALSE;
  BOOL bEnforceMaxRead = lpdwMaxRead != NULL;
  char cNext;
  *lpdwLineLen = 0;
  if (lpdwMaxRead) {
    if (*lpdwMaxRead == 0)
      return FALSE;
  }
  do
  {
    /* Determine if the string allocation needs an extra character */
    if ((*lpdwLineLen) + 2 > *lpdwLineAlloc) {
      LPTSTR lpszOldLine = *lpszLine;
      *lpdwLineAlloc = (*lpdwLineLen) + 32 * sizeof(char);
      *lpszLine = realloc(*lpszLine, *lpdwLineAlloc);

      /* If the allocation failed, create a blank string and return FALSE */
      if (!(*lpszLine)) {
        free(lpszOldLine);
        *lpszLine = malloc(sizeof(char));
        (*lpszLine)[0] = 0;
        *lpdwLineLen = 0;
        *lpdwLineAlloc = sizeof(char);
        return FALSE;
      }
    }

    /* Read in the next character */
    if (fread(&cNext, 1, 1, pInput) < 1)
      bExit = TRUE;
    else
    {
      /* If this character is a not a newline or a carriage return, then add it */
      if (bEnforceMaxRead)
        (*lpdwMaxRead)--;
      if ((cNext != '\n') && (cNext != '\r'))
        (*lpszLine)[(*lpdwLineLen)++] = cNext;
    }
  } while (!bExit && (cNext != '\n') && ((bEnforceMaxRead)? *lpdwMaxRead: TRUE));

  /* Terminate the string */
  (*lpszLine)[*lpdwLineLen] = 0;

  /* Return whether or not more data probably exists in the file */
  return !bExit;
}
