/*
Directory listing CGI program
By Steven Lawrance
(c) Steven Lawrance, 2000, 2002, 2004
(c) Bryant College, 1998-2000

Permission granted for release under the GPL in Q2 2000
by Bryant College.  Steven Lawrance can be contacted at
slawrance@yahoo.com.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "dirlist.h"
#include "globals.h"
#include "util.h"
#include "io.h"

/* Trims the leading and/or trailing whitespace in the string, if any */
LPTSTR TrimString(LPTSTR str) {
  DWORD start;
  DWORD dwSize;
  DWORD i;

  /* Check for a NULL string */
  if (!str)
    return NULL;

  /* Get the string's length */
  dwSize = strlen(str);

  /* If we're a blank string, then don't do anything */
  if (dwSize == 0)
    return str;
  i = dwSize - 1;

  /* Remove all whitespace at the end, if any */
  while (str[i] == ' ') {
    str[i] = 0;
  dwSize--;
    if (i >= 1)
      i--;
    else
      break;
  }

  /* Find the end of any whitespace at the beginning */
  start = 0;
  while (str[start] == ' ')
    start++;

  /* Move the string backwards, if needed */
  if (start) {
    for (i = 0; i < dwSize - start; i++)
      str[i] = str[i + start];
    str[i] = 0;
  }

  /* Return the string */
  return str;
}

/* Print an informational message to stdout.  If bAuditEnable is set, then add the text to the audit buffer */
DWORD infprintf(LPCTSTR args, ...)
{
  DWORD dwSize;
  va_list params;
  va_start(params, args);
  if (bAuditEnable)
  {
    /* Format the message into a string */
    LPTSTR lpszOldBuffer;
#ifdef HAVE_VSNPRINTF
    /* Safer argument parser than vsprintf(). Code adapted from GNU's vsnprintf() man page.
       This code grows the output string to a size that can accommodate the formatted string */
    int size = 4096;
    LPTSTR output = (LPTSTR) malloc(size);
    LPTSTR oldOutput;
    int n;
    if (!output) {
      printf("-Out of memory in infprintf() 1\n");
      va_end(params);
      return 0;
    }
    while (1) {
      n = vsnprintf(output, size, args, params);
      va_end(params);
      if (n > -1 && n < size)
        break;        /* vsnprintf() succeeded */
      if (n > -1)     /* glibc 2.1 */
        size = n + 1; /* precisely what is needed */
      else            /* glibc 2.0 */
        size *= 2;    /* twice the old size */
      oldOutput = output;
      if ((output = realloc(output, size)) == NULL) {
        free(oldOutput);
        printf("-Out of memory in infprintf() 2\n");
        return 0;
      }
      va_start(params, args);
    }
    dwSize = (DWORD) n;
#else
#warning Using vsprintf() instead of vsnprintf() in infprintf(), which is less secure. Ensure vsnprintf() is on your system and re-run configure to be more secure.
    LPTSTR output = (LPTSTR) malloc(65536); /* hopefully that's enough ;-) */
    if (!output) {
      printf("-Out of memory in infprintf() 3\n");
      va_end(params);
      return 0;
    }
    vsprintf(output, args, params);
    va_end(params);
    dwSize = strlen(output);
#endif

    /* Increase the size of the audit buffer (or allocate if it doesn't exist) */
    if (!lpszAuditBuffer) {
      lpszOldBuffer = NULL;
      lpszAuditBuffer = (LPTSTR) malloc(dwSize + 1);
    } else {
      lpszOldBuffer = lpszAuditBuffer;
      lpszAuditBuffer = (LPTSTR) realloc(lpszAuditBuffer, dwAuditBufferSize + dwSize + 1);
    }

    /* If we could not allocate the buffer, then return zero bytes written */
    if (!lpszAuditBuffer) {
      free(output);
      lpszAuditBuffer = lpszOldBuffer;
      return 0;
    }

    /* Concatenate the strings (use strcpy instead of strcat for increased speed) */
    strcpy(&(lpszAuditBuffer[dwAuditBufferSize]), output);

    /* Increase dwAuditBufferSize */
    dwAuditBufferSize += dwSize;

    /* Free the string after writing it to stdout */
    printf("%s", output);
    free(output);
  } else
  {
    /* Write to stdout */
    dwSize = vprintf(args, params);
    va_end(params);
  }
  return dwSize;
}

/* Replaces all non alphanumeric/period/@/-/_characters with a period and returns the
length of the string */
DWORD SafeEMailAddress(LPTSTR str)
{
  /* Get the length */
  DWORD dwLen = strlen(str);
  DWORD i;
  char current;

  /* Convert non-alphanumeric/period/@/-/_ characters to . */
  for (i = 0; i < dwLen; i++) {
    current = str[i];
    if (!((current >= 'a' && current <= 'z') || (current >= 'A' && current <= 'Z') ||
          (current >= '0' && current <= '9') || (current == '.') || (current == '@') ||
          (current == '-') || (current == '_')))
      str[i] = '.';
  }

  /* Return the length */
  return dwLen;
}

/* Replaces all non alphanumeric/period characters with an X and returns
the length of the string */
DWORD SafeDomainName(LPTSTR str)
{
  /* Get the length */
  DWORD dwLen = strlen(str);
  DWORD i;
  char current;

  /* Convert non-alphanumeric characters to Xes */
  for (i = 0; i < dwLen; i++) {
    current = str[i];
    if (!((current >= 'a' && current <= 'z') || (current >= 'A' && current <= 'Z') ||
          (current >= '0' && current <= '9') || (current == '.') || (current == '-' && i)))
      str[i] = 'X';
  }

  /* Return the length */
  return dwLen;
}

/* Replaces all non alphanumeric characters with a space and returns the
length of the string */
DWORD SafeFileName(LPTSTR str)
{
  /* Get the length */
  DWORD dwLen = strlen(str);
  DWORD i;
  char current;

  /* Convert non-alphanumeric characters to spaces */
  for (i = 0; i < dwLen; i++) {
    current = str[i];
    if (!((current >= 'a' && current <= 'z') || (current >= 'A' && current <= 'Z') ||
          (current >= '0' && current <= '9') || (current == ' ')))
      str[i] = ' ';
  }

  /* Return the length */
  return dwLen;
}

/* Compare two strings using a LIKE method */
BOOL SearchLIKEStrings(LPCTSTR str, LPCTSTR startstr, BYTE cEscape) {
  register UCHAR cmpChar1, cmpChar2;
  BOOL bFound = TRUE;
  DWORD i = 0, t = 0;

/*

If cEscape exists, skip ahead one character
If % is found and bEscape is false, find the following substring (if any)
If _ is found and bEscape is false, assume a match exists, otherwise, compare the characters

If the substring was not found, a match does not exist
If the substring was found, resume normal operations

*/

  /* Ensure the strings are not NULL */
  if (!str || !startstr)
    return (BOOL) NULL;

  /* Enter the comparison loop */
  while (bFound && (str[i] != 0) && (startstr[t] != 0))
  {
    /* Check to see if this is an escape character; skip ahead if so */
    if (startstr[t] == cEscape)
      t++;
    else
    {
      /* If this is %, find the next substring */
      if (startstr[t] == '%')
      {
        /* Get the next substring */
        if (startstr[t + 1])
        {
          DWORD t_seek = t + 1, i_seek = i, i_iteration = 0, t_orig;
          while (startstr[t_seek] && str[i_seek] && (startstr[t_seek] != '%'))
          {
            /* If this is an escape, seek ahead one more */
            t_orig = t_seek;
            if (startstr[t_seek] == cEscape)
              t_seek++;

            /* If these characters match, resume the seek; otherwise, reset */
            cmpChar1 = str[i_seek];
            cmpChar2 = startstr[t_seek];
            if ((cmpChar1 >= 'a') && (cmpChar1 <= 'z'))
              cmpChar1 -= 'a' - 'A';
            if ((cmpChar2 >= 'a') && (cmpChar2 <= 'z'))
              cmpChar2 -= 'a' - 'A';
            if ((startstr[t_orig] == '_') || (cmpChar1 == cmpChar2)) {
              t_seek++;
              i_seek++;
            } else {
              i_seek = i + (++i_iteration);
              t_seek = t + 1;
            }
          }

          /* If we could not find the substring, then this string doesn't match */
          if (startstr[t_seek] && (startstr[t_seek] != '%'))
            bFound = FALSE;

          /* Set the new i and t, and continue */
          i = i_seek;
          t = t_seek;
          continue;
        } else
        {
          /* The rest of the string matches ('%' was at the end) */
          t++;
          while (str[i])
            i++;
          continue;
        }
      } else if (startstr[t] == '_')
      {
        /* This is a match unconditionally */
        t++;
        i++;
        continue;
      }
    }

    /* Compare the current characters */
    cmpChar1 = str[i];
    cmpChar2 = startstr[t];
    if ((cmpChar1 >= 'a') && (cmpChar1 <= 'z'))
      cmpChar1 -= 'a' - 'A';
    if ((cmpChar2 >= 'a') && (cmpChar2 <= 'z'))
      cmpChar2 -= 'a' - 'A';
    if (cmpChar1 != cmpChar2)
      bFound = FALSE;
    else {
      i++;
      t++;
    }
  }

  /* Use length as a deciding factor if needed */
  if (bFound) {
    if ((str[i] == 0) && (startstr[t] == '%'))
      t++;
    if ((str[i] == 0) && (startstr[t] != 0))
      bFound = FALSE;
    if (str[i])
      bFound = FALSE;
  }
  return bFound;
}

/* Compare two strings using an "ends with" method */
BOOL SearchEndingStrings(LPCTSTR str, LPCTSTR startstr) {
  register UCHAR cmpChar1, cmpChar2;
  BOOL bFound = TRUE;
  DWORD i, t;

  /* Ensure the strings are not NULL */
  if (!str || !startstr)
    return (BOOL) NULL;

  /* Check the lengths */
  i = strlen(str);
  t = strlen(startstr);
  if (!i && t)
    return FALSE;
  if (!t)
    return TRUE;

  /* Compare the string contents */
  do {
    i--;
    t--;
    cmpChar1 = str[i];
    cmpChar2 = startstr[t];
    if ((cmpChar1 >= 'a') && (cmpChar1 <= 'z'))
      cmpChar1 -= 'a' - 'A';
    if ((cmpChar2 >= 'a') && (cmpChar2 <= 'z'))
      cmpChar2 -= 'a' - 'A';
    if (cmpChar1 != cmpChar2)
      bFound = FALSE;
  } while (bFound && i && t);

  /* Use length as a deciding factor if needed */
  if (bFound) {
    if (!i && t)
      bFound = FALSE;
  }
  return bFound;
}

/* Compare two strings using a "starts with" method */
BOOL SearchStartingStrings(LPCTSTR str, LPCTSTR startstr) {
  register UCHAR cmpChar1, cmpChar2;
  BOOL bFound = TRUE;
  DWORD i = 0;

  /* Ensure the strings are not NULL */
  if (!str || !startstr)
    return (BOOL) NULL;

  /* Compare the string contents */
  while (bFound && (str[i] != 0) && (startstr[i] != 0)) {
    cmpChar1 = str[i];
    cmpChar2 = startstr[i];
    if ((cmpChar1 >= 'a') && (cmpChar1 <= 'z'))
      cmpChar1 -= 'a' - 'A';
    if ((cmpChar2 >= 'a') && (cmpChar2 <= 'z'))
      cmpChar2 -= 'a' - 'A';
    if (cmpChar1 != cmpChar2)
      bFound = FALSE;
    else
      i++;
  }

  /* Use length as a deciding factor if needed */
  if (bFound) {
    if ((str[i] == 0) && (startstr[i] != 0))
      bFound = FALSE;
  }
  return bFound;
}

/* Compare two strings using a very strict "starts with" method (produces smaller lists) */
BOOL SearchStrictStartingStrings(LPCTSTR str, LPCTSTR startstr) {
  register UCHAR cmpChar1, cmpChar2;
  BOOL bFound = TRUE;
  DWORD i = 0;

  /* Ensure the strings are not NULL */
  if (!str || !startstr)
    return (BOOL) NULL;

  /* Compare the string contents */
  while (bFound && (str[i] != 0) && (startstr[i] != 0)) {
    cmpChar1 = str[i];
    cmpChar2 = startstr[i];
    if ((cmpChar1 >= 'a') && (cmpChar1 <= 'z'))
      cmpChar1 -= 'a' - 'A';
    if ((cmpChar2 >= 'a') && (cmpChar2 <= 'z'))
      cmpChar2 -= 'a' - 'A';
    if (cmpChar1 != cmpChar2)
      bFound = FALSE;
    else
      i++;
  }

  /* Use length as a deciding factor if needed (string pos MUST have ' ' or NULL) */
  if (bFound) {
    if ((str[i] == 0) && (startstr[i] != 0))
      bFound = FALSE;
    if ((str[i] != ' ') && (str[i] != 0))
      bFound = FALSE;
  }
  return bFound;
}

/* Compare two strings */
int CompareStrings(LPCTSTR str1, LPCTSTR str2) {
  register UCHAR cmpChar1, cmpChar2;
  int equality = 0;
  int i = 0;

  /* Ensure the strings are not NULL */
  if (!str1) {
    if (!str2)
      return 0;
    else
      return -1;
  }
  if (!str2)
    return 1;

  /* Compare the string contents */
  while ((equality == 0) && (str1[i] != 0) && (str2[i] != 0)) {
    cmpChar1 = str1[i];
    cmpChar2 = str2[i];
    if ((cmpChar1 >= 'a') && (cmpChar1 <= 'z'))
      cmpChar1 -= 'a' - 'A';
    if ((cmpChar2 >= 'a') && (cmpChar2 <= 'z'))
      cmpChar2 -= 'a' - 'A';
    if (cmpChar1 > cmpChar2)
      equality = 1;
    else {
      if (cmpChar1 < cmpChar2)
        equality = -1;
      i++;
    }
  }

  /* Use length as a deciding factor if needed */
  if (equality == 0) {
    if ((str1[i] == 0) && (str2[i] != 0))
      equality = -1;
    else {
      if ((str1[i] != 0) && (str2[i] == 0))
        equality = 1;
    }
  }
  return equality;
}

/* Find a sub-string */
BOOL FindSubstring(LPCTSTR str, LPCTSTR substr) {
  BOOL bFound = FALSE;
  register char cmpChar1, cmpChar2;
  int i = 0, len = strlen(str), t, sublen = strlen(substr);

  /* Ensure the strings are not NULL */
  if (!str) {
    if (!substr)
      return TRUE;
    else
      return FALSE;
  }
  if (!substr)
    return TRUE;

  /* Find the location of the substring in the string */
  while ((i < len) && !bFound)
  {
    /* Secondary loop for comparing the substring */
    t = 0;
    while (((i + t) < len) && (t < sublen)) {
      cmpChar1 = str[i + t];
      cmpChar2 = substr[t];
      if ((cmpChar1 >= 'a') && (cmpChar1 <= 'z'))
        cmpChar1 -= 'a' - 'A';
      if ((cmpChar2 >= 'a') && (cmpChar2 <= 'z'))
        cmpChar2 -= 'a' - 'A';
      if (cmpChar1 == cmpChar2)
        t++;
      else
        break;
    }

    /* If we reached the end of the substring, then we found it :-)! */
    if (t == sublen)
      bFound = TRUE;

    /* Next position */
    i++;
  }

  /* Return whether or not we found it :-) */
  return bFound;
}

/* Compare two names */
int CompareNames(LPCTSTR lpszFirst1, LPCTSTR lpszLast1, LPCTSTR lpszFirst2, LPCTSTR lpszLast2, BOOL bLastFirst) {
  int equality = 0;

  /* If any strings are NULL, no comparison can be made (negative 2 is returned) */
  if (!lpszLast1 || !lpszLast2 || !lpszFirst1 || !lpszFirst2)
    return -2;

  /* Sort as last name first */
  if (bLastFirst)
  {
    /* Both have two names */
    if ((lpszLast1[0] != 0) && (lpszLast2[0] != 0))
    {
      /* Compare the last name */
      equality = CompareStrings(lpszLast1, lpszLast2);

      /* If the last names equal, then do the first names */
      if (equality == 0)
        equality = CompareStrings(lpszFirst1, lpszFirst2);
    }

    /* The first has a last name, but the second does not */
    if ((lpszLast1[0] != 0) && (lpszLast2[0] == 0))
    {
      /* Compare the last name */
      equality = CompareStrings(lpszLast1, lpszFirst2);

      /* If the last names equal, then the second name is first */
      if (equality == 0)
        equality = 1;
    }

    /* The last has a last name names, but the first does not */
    if ((lpszLast1[0] == 0) && (lpszLast2[0] != 0))
    {
      /* Compare the last name */
      equality = CompareStrings(lpszFirst1, lpszLast2);

      /* If the last names equal, then the first name is first */
      if (equality == 0)
        equality = -1;
    }

    /* If neither have last names, then simply compare the first names */
    if ((lpszLast1[0] == 0) && (lpszLast2[0] == 0))
      equality = CompareStrings(lpszFirst1, lpszFirst2);
  } else
  {
    /* Compare the first names */
    equality = CompareStrings(lpszFirst1, lpszFirst2);

    /* If they are equal, then do the last names */
    if (equality == 0)
      equality = CompareStrings(lpszLast1, lpszLast2);
  }

  /* Return the equality */
  return equality;
}

BOOL ProcessWildcards(LPCTSTR file, LPCTSTR search, int searchLastPeriod, int searchExtensionLength, BOOL caseSensitive) {
  int fileLastPeriod, fileExtensionLength;
  int fileLength, searchLength;
  BOOL found = FALSE;
  LPTSTR temp;
  UCHAR comp1, comp2;
  int i, t;

  /* If there's no search, then return true */
  if (search == NULL)
    return TRUE;

  /* Get the string lengths */
  fileLength = strlen(file);
  searchLength = strlen(search);

  /* Initialize the search variables */
  temp = strrchr(file, '.');
  if (temp) {
    fileExtensionLength = strlen(temp + 1);
    fileLastPeriod = (int) (temp - file);
  } else {
    fileExtensionLength = 0;
    fileLastPeriod = -1;
  }

  /* Make sure the extensions are ok */
  if ((fileExtensionLength == 0) && (searchExtensionLength != 0))
    return FALSE;

  /* If we have an extension, then process it */
  if (searchExtensionLength) {
    /* Ensure the correctness of the extensions */
    i = searchLastPeriod + 1;
    t = fileLastPeriod + 1;
    while ((i < searchLength) && (t < fileLength) && (!found)) {
      switch (search[i]) {
      case '*' :
        found = TRUE;
        break;

      case '?' :
        i++;
        t++;
        break;

      default :
        comp1 = search[i];
        comp2 = file[t];
        if (!caseSensitive && (search[i] < 'a'))
          comp1 = (UCHAR) (search[i] + ('a' - 'A'));
        if (!caseSensitive && (file[t] < 'a'))
          comp2 = (UCHAR) (file[t] + ('a' - 'A'));
        if (comp1 == comp2) {
          i++;
          t++;
        } else
          found = -1;
        break;
      }
    }
    if (found == -1)
      return FALSE;

    /* If the search extension ends with an * and everything */
    /* was spiffy, then we found it :-)!                     */
    if ((!found) && (i < searchLength))
      if (search[i] == '*')
        found = TRUE;

      /* Extensions did not have the same length */
      if (!found && ((fileLastPeriod + 1 + (searchLength - (searchLastPeriod + 1))) != fileLength))
        return FALSE;
  } else {
    /* Set the lastPeriod variables to the end, because we */
    /* didn't have a last period!                          */
    searchLastPeriod = searchLength;
    fileLastPeriod = fileLength;
  }

  /* So far, so good...  Now let's check the filename itself */
  found = FALSE;
  i = 0;
  t = 0;
  while ((i < searchLastPeriod) && (t < fileLastPeriod) && (!found)) {
    switch (search[i]) {
    case '*' :
      found = TRUE;
      break;

    case '?' :
      i++;
      t++;
      break;

    default :
      comp1 = search[i];
      comp2 = file[t];
      if (!caseSensitive && (search[i] < 'a'))
        comp1 = (UCHAR) (search[i] + ('a' - 'A'));
      if (!caseSensitive && (file[t] < 'a'))
        comp2 = (UCHAR) (file[t] + ('a' - 'A'));
      if (comp1 == comp2) {
        i++;
        t++;
      } else
        found = -1;
      break;
    }
  }
  if (found == -1)
    return FALSE;

  /* If the search filename ends with an * and everything */
  /* was spiffy, then we found it :-)!                    */
  if ((!found) && (i < searchLastPeriod)) {
    if (search[i] == '*')
      found = TRUE;
  }

  /* Filenames did not have the same length */
  if (!found && (searchLastPeriod != fileLastPeriod))
    return FALSE;

  /* Woah!  It passed the test!!!!  Horray :-)! */
  return TRUE;
}

BOOL FindFirst(LPCTSTR lpszDir, LPCTSTR lpszName, LPFILEFIND_DATA ffData, BOOL caseSensitive) {
  /* Find the first file */
  DIR *dir;
  struct dirent *ent;
  int searchLastPeriod;
  int searchExtensionLength;
  LPTSTR temp;

  /* Initialize the search variables, if any */
  if (lpszName) {
    temp = strrchr(lpszName, '.');
    if (temp) {
      searchExtensionLength = strlen(temp + 1);
      searchLastPeriod = (int) (temp - lpszName);
    }  else {
      searchExtensionLength = 0;
      searchLastPeriod = -1;
    }
  } else {
    searchExtensionLength = 0;
    searchLastPeriod = 0;
  }

  /* Open the directory */
  if ((dir = opendir(lpszDir)) != NULL) {
    /* Scan directory until we reach the end or find the file, */
    /* whichever comes first                                   */
    while ((ent = readdir(dir)) != NULL) {
      if (ProcessWildcards(ent->d_name, lpszName, searchLastPeriod, searchExtensionLength, caseSensitive)) {
        /* We found it!  Initialize the structure */
        DIR *testDir;
        LPTSTR testDirName;
        ffData->ff_name = (LPTSTR) malloc(strlen(ent->d_name) + 1);
        ffData->ff_dir = (LPTSTR) malloc(strlen(lpszDir) + 1);
        if (lpszName)
          ffData->ff_find = (LPTSTR) malloc(strlen(lpszName) + 1);
        else
          ffData->ff_find = NULL;
        ffData->searchLastPeriod = searchLastPeriod;
        ffData->searchExtensionLength = searchExtensionLength;
        ffData->caseSensitive = caseSensitive;
        strcpy(ffData->ff_name, ent->d_name);
        strcpy(ffData->ff_dir, lpszDir);
        if (lpszName)
          strcpy(ffData->ff_find, lpszName);

        /* Determine if this is a directory */
        testDirName = (LPTSTR) malloc(strlen(lpszDir) + strlen(ent->d_name) + 2);
        sprintf(testDirName, "%s%c%s", lpszDir, cDirDelimiter, ent->d_name);
        testDir = opendir(testDirName);
        if (testDir != NULL) {
          closedir(testDir);
          ffData->ff_attrib = FA_DIREC;
        } else
          ffData->ff_attrib = FA_FILE;
        free(testDirName);

        /* Set the opened directory value and return true */
        ffData->openDir = dir;
        return TRUE;
      }
    }

    /* Close the directory and return false */
    closedir(dir);
    return FALSE;
  } else
    /* Directory not found; return false */
    return FALSE;
}

BOOL FindNext(LPFILEFIND_DATA ffData) {
  /* Find the next file */
  DIR *dir = ffData->openDir;
  struct dirent *ent;
  LPTSTR lpszName = ffData->ff_find;
  LPTSTR lpszDir = ffData->ff_dir;
  int searchLastPeriod = ffData->searchLastPeriod;
  int searchExtensionLength = ffData->searchExtensionLength;
  BOOL caseSensitive = ffData->caseSensitive;

  /* Scan directory until we reach the end or find the file, */
  /* whichever comes first                                   */
  while ((ent = readdir(dir)) != NULL) {
    if (ProcessWildcards(ent->d_name, lpszName, searchLastPeriod, searchExtensionLength, caseSensitive)) {
      /* We found it!  Initialize the structure */
      DIR *testDir;
      LPTSTR testDirName;
      free(ffData->ff_name);
      ffData->ff_name = (LPTSTR) malloc(strlen(ent->d_name) + 1);
      strcpy(ffData->ff_name, ent->d_name);

      /* Determine if this is a directory */
      testDirName = (LPTSTR) malloc(strlen(lpszDir) + strlen(ent->d_name) + 2);
      sprintf(testDirName, "%s%c%s", lpszDir, cDirDelimiter, ent->d_name);
      testDir = opendir(testDirName);
      if (testDir != NULL) {
        closedir(testDir);
        ffData->ff_attrib = FA_DIREC;
      } else
        ffData->ff_attrib = FA_FILE;
      free(testDirName);

      /* Return true */
      return TRUE;
    }
  }

  /* Close the directory and return false */
  closedir(dir);
  ffData->openDir = NULL;
  return FALSE;
}

void CloseFind(LPFILEFIND_DATA ffData) {
  /* Finish the search by freeing the memory in the structure */
  if (ffData->openDir)
    closedir(ffData->openDir);
  if (ffData->ff_name)
    free(ffData->ff_name);
  if (ffData->ff_dir)
    free(ffData->ff_dir);
  if (ffData->ff_find)
    free(ffData->ff_find);
}


/* Verifies a HTML file as valid or invalid based on its header tag */
BOOL VerifyHTML(FILE *pFile) {
  FILEFIND_DATA ffData;
  BOOL valid = FALSE;
  LPTSTR line = malloc(32768); /* 32k line limits */
  DWORD len, i;
  int num;
  line[32767] = 0;
  memset(&ffData, (char) NULL, sizeof(FILEFIND_DATA));
  while (!valid && (fgets(line, 32767, pFile) != NULL)) {
    /* Remove the upper cases in this line */
    len = strlen(line);
    for (i = 0; i < len; i++) {
      if ((line[i] >= 'A') && (line[i] <= 'Z'))
        line[i] += ('a' - 'A');
    }

    /* See if the <html tag appears */
    if (strstr(line, "<html"))
      valid = TRUE;
  }
  free(line);

  /* Invalid HTML file; return FALSE */
  if (!valid) {
    fclose(pFile);
    return FALSE;
  }

  /* Check this HTML file against our "bad templates," if any */
  if (FindFirst("dirlist_html", "badtemplate*.html", &ffData, TRUE)) {
    do {
      sscanf(ffData.ff_name, "badtemplate%i.html", &num);
      SeekFile(pFile, 0, SEEK_SET);
      if (!CheckBadTemplate(pFile, num))
        valid = FALSE;
    } while (valid && FindNext(&ffData));
  }
  CloseFind(&ffData);

  /* Return our success */
  fclose(pFile);
  return valid;
}

/* Create the /etc/passwd cache (UNIX only) */
void CreatePasswdCache()
{
  /* Fail if the delimiter is not '/' */
  FILE *pwf;
  LPTSTR lpszLine;
  LPTSTR lpszUser;
  LPDIRLIST_PASSWD_ENTRY lpCurrent;
  LPDIRLIST_PASSWD_ENTRY lpNew = NULL;
  BOOL bFound;
  DWORD i;
  long int size;
  if (cDirDelimiter != '/')
    return;

  /* Open the file */
  if (!(pwf = fopen("/etc/passwd", "r")))
    return;

  /* Allocate a string the size of the file */
  fseek(pwf, 0, SEEK_END);
  if (!(size = ftell(pwf))) {
    fclose(pwf);
    return;
  }
  if (!(lpszPasswdCache = (LPTSTR) malloc(size + 1))) {
    fclose(pwf);
    return;
  }

  /* Read the file into memory */
  fseek(pwf, 0, SEEK_SET);
  if (fread(lpszPasswdCache, 1, size, pwf) != size) {
    free(lpszPasswdCache);
    lpszPasswdCache = NULL;
    fclose(pwf);
    return;
  }
  fclose(pwf);
  lpszPasswdCache[size] = 0;

  /* Now create the binary tree */
  lpszLine = strtok(lpszPasswdCache, "\n");
  do
  {
    /* Get the username */
    lpszUser = strtok2(lpszLine, ':');

    /* Create an allocation */
    if (!lpPasswdCache)
      lpNew = lpPasswdCache = (LPDIRLIST_PASSWD_ENTRY) malloc(sizeof(DIRLIST_PASSWD_ENTRY));
    else
    {
      /* Find a blank spot */
      lpCurrent = lpPasswdCache;
      bFound = FALSE;
      while (!bFound) {
        if (CompareStrings(lpszUser, lpCurrent->lpszUser) <= 0) {
          if (!lpCurrent->left) {
            lpNew = (LPDIRLIST_PASSWD_ENTRY) malloc(sizeof(DIRLIST_PASSWD_ENTRY));
            lpCurrent->left = lpNew;
            bFound = TRUE;
          } else
            lpCurrent = lpCurrent->left;
        } else {
          if (!lpCurrent->right) {
            lpNew = (LPDIRLIST_PASSWD_ENTRY) malloc(sizeof(DIRLIST_PASSWD_ENTRY));
            lpCurrent->right = lpNew;
            bFound = TRUE;
          } else
            lpCurrent = lpCurrent->right;
        }
      }
    }

    /* Set the fields */
    lpNew->lpszUser = lpszUser;
    for (i = 0; i < 3; i++)
      strtok2(NULL, ':');
    lpNew->lpszName = strtok2(NULL, ':');

    /* Set the left and right fields to NULL */
    lpNew->left = NULL;
    lpNew->right = NULL;
  } while (lpszLine = strtok(NULL, "\n"));
}

/* Fetch an entry from /etc/passwd (_really_ fast if /etc/passwd cache is on :-) */
BOOL GetPasswdEntry(LPCTSTR lpszUser, LPTSTR *lpszName)
{
  /* If the cache exists, then use it */
  if (lpPasswdCache)
  {
    BOOL bFound = FALSE;
    LPDIRLIST_PASSWD_ENTRY lpCurrent = lpPasswdCache;
    int result;

    /* Traverse the binary tree */
    while (!bFound && lpCurrent) {
      result = CompareStrings(lpszUser, lpCurrent->lpszUser);
      if (result == 0) {
        *lpszName = lpCurrent->lpszName;
        bFound = TRUE;
      } else if (result < 0)
        lpCurrent = lpCurrent->left;
      else
        lpCurrent = lpCurrent->right;
    }

    /* If we got it, return TRUE */
    if (bFound) {
      strtok(*lpszName, ",");
      return TRUE;
    } else
      return FALSE;
  } else
  {
    /* Use the traditional getpwnam() function */
    struct passwd *pw;
    pw = getpwnam(lpszUser);
    if (pw != NULL) {
      if ((*lpszName = strtok(pw->pw_gecos, ",")) == NULL)
        *lpszName = pw->pw_gecos;
      return TRUE;
    } else
      return FALSE;
  }
}

/* Used by FreePasswdCache() */
void FreePasswdCacheNode(LPDIRLIST_PASSWD_ENTRY lpNode)
{
  /* Free children */
  if (lpNode->left) {
    FreePasswdCacheNode(lpNode->left);
    free(lpNode->left);
  }
  if (lpNode->right) {
    FreePasswdCacheNode(lpNode->right);
    free(lpNode->right);
  }
}

/* Free the /etc/passwd cache */
void FreePasswdCache()
{
  /* Free everything if we have something */
  if (lpPasswdCache)
  {
    FreePasswdCacheNode(lpPasswdCache);
    free(lpPasswdCache);
    lpPasswdCache = NULL;
    free(lpszPasswdCache);
    lpszPasswdCache = NULL;
  }
}

/* Our version of the strtok function */
LPTSTR strtok2(LPTSTR lpszStr, UCHAR cDelimiter)
{
  /* Variables */
  static LPTSTR lpszLast;
  static DWORD dwNext;
  static DWORD dwLen;
  LPTSTR lpszNext;

  /* If we have a new string, then set lpszLast */
  if (lpszStr) {
    lpszLast = lpszStr;
    dwLen = strlen(lpszStr);
    dwNext = 0;
  }

  /* Get the next delimiter if in range */
  if (dwNext <= dwLen) {
    DWORD dwLoc;
    lpszNext = strchr(&(lpszLast[dwNext]), cDelimiter);
    if (!lpszNext)
    {
      /* This is the last one */
      dwLoc = dwNext;
      dwNext = dwLen + 1;
      return &(lpszLast[dwLoc]);
    }

    /* Set dwNext equal to the index of the next delimiter */
    dwLoc = dwNext;
    dwNext = (DWORD) (((P_TYPE) lpszNext - (P_TYPE) lpszLast + 1) & 0xFFFFFFFF);

    /* Chop the string by changing the delimiter to a terminating NULL */
    lpszLast[dwNext - 1] = 0;

    /* Return the current string */
    return &(lpszLast[dwLoc]);
  }

  /* No more left; return NULL */
  return NULL;
}

BOOL IsFlagSet(LPTSTR *lpszCharList, LPTSTR lpszFlagName)
{
  /* Is the requested flag set in the command line? */
  DWORD i = 1;
  BOOL bFound = FALSE, bInParamList = TRUE;
  while (lpszCharList[i] && !bFound && bInParamList)
  {
    if (lpszCharList[i][0] != '-')
      bInParamList = FALSE;
    else
    {
      if (CompareStrings(lpszCharList[i], lpszFlagName) == 0)
        bFound = TRUE;
      else
        i++;
    }
  }
  return bFound;
}

LPTSTR GetCmdLineParam(LPTSTR *lpszCharList, DWORD dwPos, LPTSTR lpszDefault)
{
  /* Get the parameter by position */

  /* Loop through the flags */
  DWORD i = 1;
  BOOL bInParamList = TRUE;
  while (lpszCharList[i] && bInParamList)
  {
    if (lpszCharList[i][0] != '-')
      bInParamList = FALSE;
    else
      i++;
  }

  /* Safely walk to the desired parameter */
  while (lpszCharList[i] && dwPos) {
    i++;
    dwPos--;
  }

  /* If we found it, then return it; otherwise, return lpszDefault */
  if (lpszCharList[i])
    return lpszCharList[i];
  else
    return lpszDefault;
}

/* Get the last name (used for importing @1 entries in the index file) */
LPTSTR GetLastName(LPCTSTR fullName)
{
  /* Get the last name by reading the elements from right to left */
  LPTSTR lastName = strrchr(fullName, ' ');
  if (!lastName)
    return NULL;
  lastName++;
/*  if (strlen(lastName) >= 5) {
    if (strncmp((LPTSTR) (lastName + 2), ". ", 2) == 0) {
      This is an initial; return the part after the initial
      return (LPTSTR) (lastName + 3);
    }
    if (lastName[2] == ' ') {
      This is an initial; return the part after the initial
      return (LPTSTR) (lastName + 2);
    }
  }*/
  return lastName;
}

/* Converts the input string into a new string using HTML-safe character escapes as
   needed. If no changes are needed, the input string is returned as-is. You must
   call free() at some point in the future on *safeStrings to free up any strings
   allocated in this function. The safeStringsLen argument contains the length
   of the string with terminating NULs included. Use STR_NUL_TERM for textLen
   if you don't know its size and it's NUL-terminated. If vaList is not NULL,
   then elements will be read from it when printf()-style formats exist in text.
   Note that printf() isn't called; this function returns a new version of the
   passed-in format string in that case with the HTML/JavaScript-safe text already
   in it */
LPCTSTR SafeWebString(LPTSTR *safeStrings, LPDWORD safeStringsLen, LPDWORD safeStringsAlloc, LPCTSTR text, DWORD textLen, DWORD escapeMode, va_list *vaList) {
  LPTSTR safeText = NULL;
  LPTSTR safeTextCurrent = NULL;
  LPCTSTR current = text;
  LPCTSTR replacement;
  DWORD replacementLen;
  BOOL inPercentEscape = FALSE;
  BOOL charCodeReplacement;
  DWORD increment;
  BOOL prependBlankStringFormat = FALSE;
  BOOL escapeSpace = FALSE;

  /* Look at each character to see if it's HTML-safe or not */
  while (*current && ((textLen != STR_NUL_TERM)? (DWORD) (current - text) < textLen: TRUE)) {

    /* Process a character for safety if this is a string and not a format string */
    replacement = NULL;
    replacementLen = 0;
    charCodeReplacement = FALSE;
    increment = 1;
    if (!vaList) {
      if (escapeMode == SAFEWEBSTRING_ESCAPE_JAVASCRIPT) {
        switch (*current) {
          case '\r' : replacement = "\\r"; replacementLen = 2; break;
          case '\n' : replacement = "\\n"; replacementLen = 2; break;
          case '\b' : replacement = "\\b"; replacementLen = 2; break;
          case '\f' : replacement = "\\f"; replacementLen = 2; break;
          case '\t' : replacement = "\\t"; replacementLen = 2; break;
          case '\v' : replacement = "\\v"; replacementLen = 2; break;
          case '"' : replacement = "\\\""; replacementLen = 2; break;
          case '\'' : replacement = "\\'"; replacementLen = 2; break;
          case '\\' : replacement = "\\\\"; replacementLen = 2; break;
          default :

            /* Generically escape everything outside of normal Latin characters */
            if (*current < ' ' || *current > '~' || *current == '<' ||
                *current == '>') {
              replacement = "\\x";
              replacementLen = 4;
              charCodeReplacement = TRUE;
            }
            break;
        }
      } else {

        /* Reset escapeSpace if this character is not a space */
        if (escapeSpace && *current != ' ') {
          escapeSpace = FALSE;
        }

        /* Handle HTML escaping */
        if (*current < ' ' || *current > '~' || (escapeMode == SAFEWEBSTRING_ESCAPE_RESPONSE &&
            (*current == '\n' || *current == '\r'))) {
          replacement = "&#";
          replacementLen = 5; /* max; ranges from 3 to 5 */
          charCodeReplacement = TRUE;
        } else {
          switch (*current) {
            case '<' : replacement = "&lt;"; replacementLen = 4; break;
            case '>' : replacement = "&gt;"; replacementLen = 4; break;
            case '"' : replacement = "&quot;"; replacementLen = 6; break;
            case '&' : replacement = "&amp;"; replacementLen = 5; break;

            case '\n' :

              /* In HTML mode, make this <br/> */
              replacement = "<br/>";
              replacementLen = 5;
              break;

            case '\r' :

              /* In HTML mode, replace carriage return with nothing */
              replacement = "";
              replacementLen = 0;
              break;

            case ' ' :

              /* Use &nbsp; for spaces that follow a space */
              if (!escapeSpace) {
                escapeSpace = TRUE;
              } else {
                escapeSpace = FALSE;
                replacement = "&nbsp;";
                replacementLen = 6;
              }
              break;
          }
        }
      }
    }

    /* If we have a vaList and this is a % character, then process it */
    else {
      if (!inPercentEscape) {
        if (*current == '%') {

          /* If we have a %% escape, then don't process it */
          if (current[1] == '%') {
            inPercentEscape = TRUE;
          } else {

            /* If this is a string, then recursively convert this argument now */
            if (current[1] == 's') {
              DWORD origSafeStringsLen = *safeStringsLen;
              LPCTSTR strArg = va_arg(*vaList, LPCTSTR);
              BOOL recursiveCallAllocatedSafeStrings = FALSE;
              if (!strArg) {
                strArg = "(null)";
              }

              /* Reserve space for %.0s, if needed and if safeStrings has been allocated.
                 The recursive call will write to the final destination for us */
              if (*safeStrings) {
                *safeStringsLen += 4;
              } else {
                recursiveCallAllocatedSafeStrings = TRUE;
              }
              replacement = SafeWebString(safeStrings, safeStringsLen, safeStringsAlloc, strArg, STR_NUL_TERM, escapeMode, NULL);

              /* If replacements were made, then perform a replacement on the format string */
              if (replacement != strArg) {

                /* If we already have safe text, then perform post-processing now and
                   set the replacement length to zero */
                if (safeText) {
                  memcpy(safeTextCurrent, "%.0s", 4);
                  safeTextCurrent = &((*safeStrings)[*safeStringsLen - 1]);
                  replacementLen = 0;
                } else {

                  /* Prepend the blank string format %.0s later, after safeText initialization */
                  replacementLen = *safeStringsLen - origSafeStringsLen - 1; /* remove term NUL char */
                  if (recursiveCallAllocatedSafeStrings) {
                    replacementLen += 4; /* reserve space for %.0s now that safeStrings is allocated */
                  }
                  prependBlankStringFormat = TRUE;
                }

                /* When processing is complete, we shall increment by 2 instead of 1 so that
                   we skip over both characters in "%s" */
                increment = 2;
              } else {
                replacement = NULL;
              }
            }

            /* Otherwise, skip over this argument */
            else {
              DWORD i = 1;
              char ch = current[i];

              /* Flag characters */
              while (ch && ch == '#' || ch == '0' || ch == '-' || ch == ' ' || ch == '+' ||
                      ch == '\'' || ch == 'I') {
                ch = current[++i];
              }

              /* Field width */
              if (ch && (('1' < ch && ch < '9') || ch == '*')) {
                BOOL expectTerm = ch == '*' && '0' < current[i + 1] && current[i + 1] < '9';
                ch = current[++i];
                while (ch && '0' < ch && ch < '9') {
                  ch = current[++i];
                }
                if (expectTerm && ch != '$') {
                  ch = 0; /* invalid; consider this text instead of a format */
                }
              }

              /* Precision */
              if (ch == '.') {
                BOOL expectTerm;
                ch = current[++i];
                expectTerm = ch == '*' && '0' < current[i + 1] && current[i + 1] < '9';
                while (ch && '0' < ch && ch < '9') {
                  ch = current[++i];
                }
                if (expectTerm && ch != '$') {
                  ch = 0; /* invalid; consider this text instead of a format */
                }
              }

              /* Length modifier */
              if (ch == 'h' || ch == 'l') {
                ch = current[++i];
                if (ch == 'h' || ch == 'l') {
                  ch = current[++i];
                }
              } else if (ch == 'L' || ch == 'q' || ch == 'j' || ch == 'z' || ch == 't') {
                ch = current[++i];
              }

              /* Conversion specifier. We skip over the argument here */
              switch (ch) {

                case 'd' :
                case 'i' :
                  va_arg(*vaList, int);
                  break;

                case 'o' :
                case 'u' :
                case 'x' :
                case 'X' :
                  va_arg(*vaList, unsigned int);
                  break;

                case 'e' :
                case 'E' :
                case 'f' :
                case 'F' :
                case 'g' :
                case 'G' :
                case 'a' :
                case 'A' :
                  va_arg(*vaList, double);
                  break;

                case 'C' :
                case 'c' :
                  va_arg(*vaList, int);
                  break;

                case 'S' :
                case 's' :
                  va_arg(*vaList, const char*);
                  break;

                case 'p' :
                  va_arg(*vaList, void *);
                  break;

                case 'n' :
                  va_arg(*vaList, int *);
                  break;
              }
            }
          }
        }
      } else {
        inPercentEscape = FALSE;
      }
    }

    /* If we have a replacement and this is the first one,
       then allocate a new string now */
    if (!safeText && replacement) {

      /* Allocate safeStrings if needed */
      DWORD processedLen;
      if (!*safeStrings) {
        LPTSTR newSafeStrings;
        DWORD newAlloc;
        if (textLen == STR_NUL_TERM) {
          textLen = strlen(text);
        }
        newAlloc = textLen + replacementLen + 1024;
        newSafeStrings = (LPTSTR) malloc(newAlloc);
        if (!newSafeStrings) {
          return "(out of memory in malloc() call in MakeSafeHTMLString())";
        }
        *safeStrings = newSafeStrings;
        *safeStringsAlloc = newAlloc;
        *safeStringsLen = 0;
      }

      /* Expand safeStrings if needed */
      else if (*safeStringsAlloc - *safeStringsLen - 10 < (DWORD) (current - text) + replacementLen) {
        LPTSTR newSafeStrings;
        DWORD newAlloc;
        if (textLen == STR_NUL_TERM) {
          textLen = strlen(text);
        }
        newAlloc = *safeStringsLen + textLen + replacementLen + 1024;
        newSafeStrings = (LPTSTR) realloc(*safeStrings, newAlloc);
        if (!newSafeStrings) {
          return "(out of memory in relloc() call 1 in MakeSafeHTMLString())";
        }
        *safeStrings = newSafeStrings;
        *safeStringsAlloc = newAlloc;
      }

      /* Set safeText */
      safeText = &((*safeStrings)[*safeStringsLen]);

      /* Copy the processed text thus far */
      processedLen = (DWORD) (current - text);
      memcpy(safeText, text, processedLen);
      *safeStringsLen += processedLen;
      safeTextCurrent = &safeText[processedLen];
    }

    /* If we have a safeText allocation, then copy the current safe
       character into it now */
    if (safeText) {

      /* Increase the allocation if we are running low */
      if (*safeStringsAlloc < *safeStringsLen + replacementLen + 10) {
        LPTSTR newSafeStrings;
        DWORD newAlloc;
        if (textLen == STR_NUL_TERM) {
          textLen = strlen(text);
        }
        newAlloc = *safeStringsLen + textLen - (DWORD) (current - text) + replacementLen + 1024;
        newSafeStrings = (LPTSTR) realloc(*safeStrings, newAlloc);
        if (!newSafeStrings) {
          return "(out of memory in relloc() call 2 in MakeSafeHTMLString())";
        }
        *safeStrings = newSafeStrings;
        *safeStringsAlloc = newAlloc;
      }

      /* If we don't have a replacement, simply copy the character */
      if (!replacement) {
        *safeTextCurrent = *current;
        safeTextCurrent = &safeTextCurrent[1];
        *safeStringsLen += 1;
      }

      else {

        /* Prepend a blank string format if we are supposed to */
        if (prependBlankStringFormat) {
          memcpy(safeTextCurrent, "%.0s", 4);
          safeTextCurrent = &safeTextCurrent[4];
          replacementLen -= 4;
          *safeStringsLen += 4;
          prependBlankStringFormat = FALSE;
        }

        /* If this replacement is a character code, then generate that */
        if (charCodeReplacement) {
          memcpy(safeTextCurrent, replacement, 2);
          safeTextCurrent = &safeTextCurrent[2];
          if (escapeMode == SAFEWEBSTRING_ESCAPE_JAVASCRIPT) {
            replacementLen = sprintf(safeTextCurrent, "%02x", (DWORD) *current);
          } else {
            replacementLen = sprintf(safeTextCurrent, "%u", (DWORD) *current);
          }
          safeTextCurrent = &safeTextCurrent[replacementLen];
          *safeStringsLen += replacementLen + 2;
        }

        /* Otherwise, copy the replacement verbatim */
        else if (replacementLen) {
          memcpy(safeTextCurrent, replacement, replacementLen);
          safeTextCurrent = &safeTextCurrent[replacementLen];
          *safeStringsLen += replacementLen;
        }
      }
    }

    /* Next character */
    current = &current[increment];
  }

  /* If we have safe HTML text, then return that. Otherwise, return the input text */
  if (safeText) {

    /* Terminate safeText and return it */
    *safeTextCurrent = '\0';
    *safeStringsLen += 1;
    return safeText;
  } else {
    return text;
  }
}

LPCTSTR SafeHTMLString(LPTSTR *safeStrings, LPDWORD safeStringsLen, LPDWORD safeStringsAlloc, LPCTSTR text, DWORD textLen) {
  return SafeWebString(safeStrings, safeStringsLen, safeStringsAlloc, text, textLen, SAFEWEBSTRING_ESCAPE_HTML, NULL);
}

LPCTSTR SafeResponseString(LPTSTR *safeStrings, LPDWORD safeStringsLen, LPDWORD safeStringsAlloc, LPCTSTR text, DWORD textLen) {
  return SafeWebString(safeStrings, safeStringsLen, safeStringsAlloc, text, textLen, SAFEWEBSTRING_ESCAPE_RESPONSE, NULL);
}

LPCTSTR SafeJavaScriptString(LPTSTR *safeStrings, LPDWORD safeStringsLen, LPDWORD safeStringsAlloc, LPCTSTR text, DWORD textLen) {
  return SafeWebString(safeStrings, safeStringsLen, safeStringsAlloc, text, textLen, SAFEWEBSTRING_ESCAPE_JAVASCRIPT, NULL);
}
