/*
Directory listing CGI program
By Steven Lawrance
(c) Steven Lawrance, 2000, 2002, 2004
(c) Bryant College, 1998-2000

Permission granted for release under the GPL in Q2 2000
by Bryant College.  Steven Lawrance can be contacted at
slawrance@yahoo.com.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "dirlist.h"
#include "web.h"
#include "database.h"
#include "io.h"
#include "util.h"
#include "globals.h"

void PrintAttribute(LPDIRLIST_ACCESSPLAN lpPlan, LPDWORD paramNum, UCHAR cDelimiter, UCHAR cStringChar, LPCTSTR lpszFormatStr, ...)
{
  /* Get the ending parameters (used when we write the inside data to the string) */
  int nPrinted;
  va_list vaList;
  va_start(vaList, lpszFormatStr);

  /* Write the row delimiter if we have one (but not if this is param zero) */
  if (cDelimiter) {
    if (*paramNum)
      putchar(cDelimiter);

    /* If this is a string, write the string delimiter */
    if (cStringChar)
      putchar(cStringChar);

    /* Write the formatted arguments to stdout */
    nPrinted = vprintf(lpszFormatStr, vaList);
    va_end(vaList);

    /* Write the closing string delimiter if we have one */
    if (cStringChar)
      putchar(cStringChar);
  } else
  {
    /* Web mode */
    LPCTSTR safeString;
    DWORD origSafeStringsLen = safeStringsLen;
    LPCTSTR safeFormatStr;

    /* Web mode; write a column tag if this isn't param zero */
    if (*paramNum) {
      if (lpPlan->bUseBackgroundColor)
        printf("</td><td bgcolor=\"#%06x\" nowrap>", lpPlan->dwBackgroundColor);
      else
        printf("</td><td nowrap>");
    }

    /* Modify the lpszFormatStr to include the HTML safe text in all %s positions */
    safeFormatStr = SafeWebString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpszFormatStr, STR_NUL_TERM, SAFEWEBSTRING_ESCAPE_HTML, &vaList);
    va_end(vaList);

    /* If the format string is blank, then write &nbsp; */
    if (!safeFormatStr[0]) {
      printf("&nbsp;");
      va_end(vaList);
    }

    /* Otherwise, write out the HTML safe text */
    else {
      va_start(vaList, lpszFormatStr);
      vprintf(safeFormatStr, vaList);
      va_end(vaList);
    }

    /* "Free" the temporary memory used, if any */
    safeStringsLen = origSafeStringsLen;
  }

  /* Increment the number of parameters processed */
  (*paramNum)++;
}

/* Display a person, version2

  lpDirEntry:   Pointer to the DIRLIST_ENTRY structure to display

  lpClassList:  Pointer to the DIRLIST_CLASS structure list with class definitions

  bAdminMode:   Administrative mode -- Bypasses permission checking (FLAGS_EXCLUDE)

  cDelimiter:   The character to use between columns, or NULL if the
                web page format is to be used

  cStringChar:  The character to print at the beginning and end of
                strings, if cDelimiter is not NULL.  If this value is
                NULL, no character is used to distinguish strings from
                other data types

  dwFields:     Fields to display, each separated by a cell boundry
  dwXFields:    Fields that, if absent from this directory entry, cause
                this person to not be written to the output web page
  dwExtFields:  Extended group IDs to display
  dwXExtFields: Extended group IDs that, if absent from this directory entry,
                cause this person to not be written to the output web page
  dwExtValues:  Extended group IDs to display as ordinal values instead of
                the string
  dwExtBoth:    Extended group IDs to display as both ordinal and string
*/
BOOL DisplayPerson(LPDIRLIST_ENTRY lpDirEntry, LPDIRLIST_ACCESSPLAN lpPlan, LPDIRLIST_CLASS lpClassList, UCHAR cDelimiter, UCHAR cStringChar)
{
  /* Initialize parameter number to zero */
  DWORD i;
  DWORD paramNum = 0;
  DWORD safeStringsLen, safeStringsAlloc;
  LPTSTR safeStrings = NULL;

  /* Check the permissions on this entry before attempting to write it */
  if (!lpPlan->bAdminMode)
  {
    /* This user MUST be hidden from all views, regardless of display mode */
    if (lpDirEntry->dwFlags & FLAGS_EXCLUDE)
      return FALSE;
  }

  /* If we are supposed to print the username, then do so before processing the name */
  if (lpPlan->dwFields & DISPLAY_USERNAME)
    PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", lpDirEntry->lpszUserName);

  /* Display the name after deciding how to print it: */
  if ((lpPlan->dwFields & DISPLAY_FIRSTNAME) || (lpPlan->dwFields & DISPLAY_LASTNAME)) {
    if (lpPlan->dwFields & DISPLAY_SEPARATENAMES)
    {
      /* Display names separately */
      /* Lastname then Firstname */
      if ((lpPlan->dwFields & DISPLAY_LASTNAMEFIRST) == DISPLAY_LASTNAMEFIRST) {
        PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", lpDirEntry->lpszLastName);
        PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", lpDirEntry->lpszFirstName);
      } else
      {
        /* Firstname then Lastname */
        PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", lpDirEntry->lpszFirstName);
        PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", lpDirEntry->lpszLastName);
      }
    } else
    {
      /* Display names in same field.  Check to see if both exist first */
      if ((lpDirEntry->lpszFirstName[0] == 0) || (lpDirEntry->lpszLastName[0] == 0))
      {
        /* Display Firstname (if both are NULL, a zero-length string gets displayed anyway :-) */
        if (lpDirEntry->lpszLastName[0] == 0)
          PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", lpDirEntry->lpszFirstName);

        /* Display Lastname */
        else
          PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", lpDirEntry->lpszLastName);
      } else
      {
        /* Firstname Lastname */
        if ((lpPlan->dwFields & DISPLAY_LASTNAME) && (lpPlan->dwFields & DISPLAY_FIRSTNAME) && !((lpPlan->dwFields & DISPLAY_LASTNAMEFIRST) == DISPLAY_LASTNAMEFIRST))
          PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s %s", lpDirEntry->lpszFirstName, lpDirEntry->lpszLastName);

        /* Lastname, Firstname */
        if ((lpPlan->dwFields & DISPLAY_LASTNAMEFIRST) == DISPLAY_LASTNAMEFIRST)
          PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s, %s", lpDirEntry->lpszLastName, lpDirEntry->lpszFirstName);

        /* Firstname */
        if (!(lpPlan->dwFields & DISPLAY_LASTNAME) && (lpPlan->dwFields & DISPLAY_FIRSTNAME))
          PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", lpDirEntry->lpszFirstName);

        /* Lastname */
        if ((lpPlan->dwFields & DISPLAY_LASTNAME) && !(lpPlan->dwFields & DISPLAY_FIRSTNAME))
          PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", lpDirEntry->lpszLastName);
      }
    }
  }

  /* Go through each attribute and process them */
  i = DISPLAY_MIN;
  while (i <= DISPLAY_MAX) {
    if (lpPlan->dwFields & i)
    {
      /* Breakout for displaying this attribute */
      switch (i) {

      case DISPLAY_FLAGS :
        if (lpPlan->bAdminMode)
          PrintAttribute(lpPlan, &paramNum, cDelimiter, 0, "%u", (lpDirEntry->dwFlags & FLAGS_EXCLUDE)? 1: 0);
        PrintAttribute(lpPlan, &paramNum, cDelimiter, 0, "%u", (lpDirEntry->dwFlags & FLAGS_NOHOMEDIR)? 1: 0);
        PrintAttribute(lpPlan, &paramNum, cDelimiter, 0, "%u", (lpDirEntry->dwFlags & FLAGS_NONAMECHECK)? 1: 0);
        PrintAttribute(lpPlan, &paramNum, cDelimiter, 0, "%u", (lpDirEntry->dwFlags & FLAGS_NOWEBSITE)? 1: 0);
        PrintAttribute(lpPlan, &paramNum, cDelimiter, 0, "%u", (lpDirEntry->dwFlags & FLAGS_NOWEBCHECK)? 1: 0);
        PrintAttribute(lpPlan, &paramNum, cDelimiter, 0, "%u", (lpDirEntry->dwFlags & FLAGS_NOEMAIL)? 1: 0);
        PrintAttribute(lpPlan, &paramNum, cDelimiter, 0, "%u", (lpDirEntry->dwFlags & FLAGS_NOTELEPHONE)? 1: 0);
        break;

      case DISPLAY_WEBSITE :
        if (((lpDirEntry->lpszWebSite[0] == 0) || (lpDirEntry->dwFlags & FLAGS_NOWEBSITE)) && !lpPlan->bAdminMode)
          PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "");
        else {
          if (cDelimiter)
            PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", lpDirEntry->lpszWebSite);
          else
            PrintAttribute(lpPlan, &paramNum, 0, 0, "<a href=\"%s\"><font name=\"ws_%s\">%s</font></a>", lpDirEntry->lpszWebSite, (lpDirEntry->lpszUserName)? lpDirEntry->lpszUserName: lpDirEntry->lpszWebSite, lpDirEntry->lpszWebSite);
        }
        break;

      case DISPLAY_EMAIL :
        if (((lpDirEntry->lpszEMail[0] == 0) || (lpDirEntry->dwFlags & FLAGS_NOEMAIL)) && !lpPlan->bAdminMode)
          PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "");
        else {
          if (cDelimiter)
            PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", lpDirEntry->lpszEMail);
          else
            PrintAttribute(lpPlan, &paramNum, 0, 0, "<a href=\"mailto:%s\"><font name=\"em_%s\">%s</font></a>", lpDirEntry->lpszEMail, lpDirEntry->lpszUserName, lpDirEntry->lpszEMail);
        }
        break;

      case DISPLAY_TELEPHONE :
        if (((lpDirEntry->lpszTelephone[0] == 0) || (lpDirEntry->dwFlags & FLAGS_NOTELEPHONE)) && !lpPlan->bAdminMode)
          PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "");
        else
          PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", lpDirEntry->lpszTelephone);
        break;
      }
    }

    /* Next attribute */
    i *= 2;
  }

  /* Display extended attributes, if any */
  i = 1;
  while (i <= MAX_CLASS)
  {
    /* If this attribute is requested for display, then get the attribute */
    if (lpPlan->dwExtFields & i) {
      LPDIRLIST_CLASSDATA lpClassData = GetClassData(lpDirEntry, i);

      /* Display the flags if requested */
      if (lpPlan->dwExtFlags)
      {
        /* If this entry has a value for this class, then display the flags */
        if (lpClassData) {
          PrintAttribute(lpPlan, &paramNum, cDelimiter, (UCHAR) NULL, "1");
          PrintAttribute(lpPlan, &paramNum, cDelimiter, (UCHAR) NULL, "%u", lpClassData->dwFlags & CLASSDATA_OTHER);
        } else
        {
          /* Entry does not have this value */
          PrintAttribute(lpPlan, &paramNum, cDelimiter, (UCHAR) NULL, "0");
          PrintAttribute(lpPlan, &paramNum, cDelimiter, (UCHAR) NULL, "0");
        }
      }

      /* Process the display of the value */
      if (lpClassData)
      {
        /* Display the value */
        if (lpPlan->dwExtBoth & i) {
          PrintAttribute(lpPlan, &paramNum, cDelimiter, (UCHAR) NULL, "%u", lpClassData->dwValue);
          PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", (lpClassData->lpszValue)? lpClassData->lpszValue: "");
        } else {
          if (lpPlan->dwExtValues & i)
            PrintAttribute(lpPlan, &paramNum, cDelimiter, (UCHAR) NULL, "%u", lpClassData->dwValue);
          else
            PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", (lpClassData->lpszValue)? lpClassData->lpszValue: "");
        }
      } else
      {
        /* Value was requested for display, but doesn't exist in this entry! */
        LPDIRLIST_CLASS lpClass = GetClass(lpClassList, i);
        if (lpClass) {
          if (lpClass->dwFlags & CLASS_DEFAULTVALUE)
          {
            /* Write the default value */
            if (lpPlan->dwExtBoth & i) {
              LPDIRLIST_CLASSVALUE lpClassValue = GetClassValue(lpClass, lpClass->idDefaultValue);
              PrintAttribute(lpPlan, &paramNum, cDelimiter, (UCHAR) NULL, "%u", lpClass->idDefaultValue);
              if (lpClassValue)
                PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", (lpClassValue->lpszValue)? lpClassValue->lpszValue: "");
              else
                PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "Could not locate default value %u in %s!", lpClass->idDefaultValue, lpClass->lpszName);
            } else {
              if (lpPlan->dwExtValues & i)
                PrintAttribute(lpPlan, &paramNum, cDelimiter, (UCHAR) NULL, "%u", lpClass->idDefaultValue);
              else {
                LPDIRLIST_CLASSVALUE lpClassValue = GetClassValue(lpClass, lpClass->idDefaultValue);
                if (lpClassValue)
                  PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "%s", (lpClassValue->lpszValue)? lpClassValue->lpszValue: "");
                else
                  PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "\n\nCould not locate default value %u in %s!\n", lpClass->idDefaultValue, lpClass->lpszName);
              }
            }
          } else {
            PrintAttribute(lpPlan, &paramNum, cDelimiter, (lpPlan->dwExtValues & i)? (UCHAR) NULL: cStringChar, "");
            if (lpPlan->dwExtBoth & i)
              PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "");
          }
        } else {
          PrintAttribute(lpPlan, &paramNum, cDelimiter, (lpPlan->dwExtValues & i)? (UCHAR) NULL: cStringChar, "");
          if (lpPlan->dwExtBoth & i)
            PrintAttribute(lpPlan, &paramNum, cDelimiter, cStringChar, "");
        }
      }
    }

    /* Increment safely */
    if (i == MAX_CLASS)
      i++;
    else
      i *= 2;
  }

  /* Write the ending of the string */
  if (cDelimiter) {
    putchar('\n');
  } else {
    if (lpPlan->bUseBackgroundColor)
      printf("</td></tr>\n<tr><td bgcolor=\"#%06x\" nowrap>", lpPlan->dwBackgroundColor);
    else
      printf("</td></tr>\n<tr><td nowrap>");
  }

  /* Entry written to stdout :-)!  Return TRUE */
  return TRUE;
}

/* Write the access plan into a JavaScript object through stdout */
void WriteAccessPlanAsJavaScript(LPDIRLIST_ACCESSPLAN lpPlan, LPTSTR lpszScriptFooter)
{
  /* Write the JavaScript escape */
  DWORD origSafeStringsLen = safeStringsLen;
  DWORD i;
  printf("<script language=\"JavaScript\"><!--\n");
  printf("// DirList2 by Steven Lawrance\n");
  printf("// Copyright (c) Steven Lawrance 2000-2007\n");
  printf("// Copyright (c) Bryant College 1998-2000\n\n");

  /* Write the search/range object definition */
  printf("function DirList2SearchRange(sf, esf, eso, sn, sa, sl, sle, se, s, rl, rh, rli, rhi) {\n");
  printf("  this.searchFields = sf;\n");
  printf("  this.extSearchFields = esf;\n");
  printf("  this.extSearchOrdinals = eso;\n");
  printf("  this.searchNot = sn;\n");
  printf("  this.searchAnd = sa;\n");
  printf("  this.searchLevel = sl;\n");
  printf("  this.searchLevelEnd = sle;\n");
  printf("  this.searchExact = se;\n");
  printf("  if (s != null) this.searchString = s; else this.searchString = \"\";\n");
  printf("  if (rl != null) this.rangeLow = rl; else this.rangeLow = \"\";\n");
  printf("  if (rh != null) this.rangeHigh = rh; else this.rangeHigh = \"\";\n");
  printf("  this.rangeLowInclusive = rli;\n");
  printf("  this.rangeHighInclusive = rhi;\n");
  printf("}\n\n");

  /* Write the toString function */
  printf("function DirList2ToString() {\n");
  printf("  str = \"\";\n");
  printf("  if (this.fields) str += \"f=\" + this.fields + \"&\";\n");
  printf("  if (this.xFields) str += \"xf=\" + this.xFields + \"&\";\n");
  printf("  if (this.extFields) str += \"ef=\" + this.extFields + \"&\";\n");
  printf("  if (this.xExtFields) str += \"xef=\" + this.xExtFields + \"&\";\n");
  printf("  if (this.extValues) str += \"ev=\" + this.extValues + \"&\";\n");
  printf("  if (this.extBoth) str += \"eb=\" + this.extBoth + \"&\";\n");
  printf("  if (this.extFlags) str += \"efl=\" + this.extFlags + \"&\";\n");
  printf("  if (this.sort) str += \"o=\" + this.sort + \"&\";\n");
  printf("  if (this.search.length) {\n");
  printf("    str += \"ns=\" + this.search.length + \"&\";\n");
  printf("    for (i = 0; i < this.search.length; i++) {\n");
  printf("      if (this.search[i].searchFields) str += \"sf\" + i + \"=\" + this.search[i].searchFields + \"&\";\n");
  printf("      if (this.search[i].extSearchFields) str += \"esf\" + i + \"=\" + this.search[i].extSearchFields + \"&\";\n");
  printf("      if (this.search[i].extSearchOrdinals) str += \"eso\" + i + \"=\" + this.search[i].extSearchOrdinals + \"&\";\n");
  printf("      if (this.search[i].searchNot) str += \"sn\" + i + \"=\" + this.search[i].searchNot + \"&\";\n");
  printf("      if (this.search[i].searchAnd) str += \"sa\" + i + \"=\" + this.search[i].searchAnd + \"&\";\n");
  printf("      if (this.search[i].searchLevel) str += \"sl\" + i + \"=\" + this.search[i].searchLevel + \"&\";\n");
  printf("      if (this.search[i].searchLevelEnd) str += \"sle\" + i + \"=\" + this.search[i].searchLevelEnd + \"&\";\n");
  printf("      if (this.search[i].searchExact) str += \"se\" + i + \"=\" + this.search[i].searchExact + \"&\";\n");
  printf("      if (this.search[i].searchString.length > 0) str += \"s\" + i + \"=\" + encodeURIComponent(this.search[i].searchString) + \"&\";\n");
  printf("      if (this.search[i].rangeLow.length > 0) str += \"rl\" + i + \"=\" + encodeURIComponent(this.search[i].rangeLow) + \"&\";\n");
  printf("      if (this.search[i].rangeHigh.length > 0) str += \"rh\" + i + \"=\" + encodeURIComponent(this.search[i].rangeHigh) + \"&\";\n");
  printf("      if (this.search[i].rangeLowInclusive) str += \"rli\" + i + \"=\" + this.search[i].rangeLowInclusive + \"&\";\n");
  printf("      if (this.search[i].rangeHighInclusive) str += \"rhi\" + i + \"=\" + this.search[i].rangeHighInclusive + \"&\";\n");
  printf("    }\n");
  printf("  }\n");
  printf("  if (this.javaScript) str += \"js=1&\";\n");
  printf("  if (this.counterBegin > 1) str += \"cb=\" + this.counterBegin + \"&\";\n");
  printf("  if (this.counterIncrement) str += \"ci=\" + this.counterIncrement + \"&\";\n");
  printf("  if (this.counterPrev) str += \"cp=\" + this.counterPrev + \"&\";\n");
  printf("  if (this.counterNext) str += \"cn=\" + this.counterNext + \"&\";\n");
  printf("  if (this.counterHints) str += \"ch=\" + this.counterHints + \"&\";\n");
  printf("  if (this.counterTotal) str += \"ct=\" + this.counterTotal + \"&\";\n");
  printf("  if (this.useBackgroundColor) str += \"ubg=1&bg=\" + this.backgroundColor + \"&\";\n");
  printf("  str += \"seq=\" + this.sequenceNumber;\n");
  printf("  return str;\n");
  printf("}\n\n");

  /* Write the header of the object */
  printf("function DirList2() {\n");

  /* Value filters */
  printf("  // Value filters for display\n");
  printf("  this.fields = %u;\n", lpPlan->dwFields);
  printf("  this.xFields = %u;\n", lpPlan->dwXFields);
  printf("  this.extFields = %u;\n", lpPlan->dwExtFields);
  printf("  this.xExtFields = %u;\n", lpPlan->dwXExtFields);
  printf("  this.extValues = %u;\n", lpPlan->dwExtValues);
  printf("  this.extBoth = %u;\n", lpPlan->dwExtBoth);
  printf("  this.extFlags = %u;\n\n", lpPlan->dwExtFlags);

  /* Sort order */
  printf("  // Sort order\n");
  printf("  this.sort = %u;\n\n", lpPlan->idSort);

  /* Search/range parameters */
  printf("  // Search/range parameters\n");
  printf("  this.search = new Array(%u);\n", lpPlan->nSearches);
  for (i = 0; i < lpPlan->nSearches; i++)
  {
    /* Create a search/range object and fill it out */
    printf("  this.search[%u] = new DirList2SearchRange(%u, %u, %u, %i, %i, %u, %u, %u, \"%s\", \"%s\", \"%s\", %i, %i);\n", i,
      lpPlan->dwSearchFields[i], lpPlan->dwExtSearchFields[i], lpPlan->dwExtSearchOrdinals[i], (lpPlan->bSearchNot[i])? TRUE: FALSE,
      (lpPlan->bSearchAnd[i])? TRUE: FALSE, lpPlan->dwSearchLevel[i], lpPlan->dwSearchLevelEnd[i],
      lpPlan->fSearchExact[i],
      SafeJavaScriptString(&safeStrings, &safeStringsLen, &safeStringsAlloc, (lpPlan->lpszRangeLow[i] == lpPlan->lpszRangeHigh[i])? lpPlan->lpszSearch[i]: "", -1),
      SafeJavaScriptString(&safeStrings, &safeStringsLen, &safeStringsAlloc, (lpPlan->lpszRangeLow[i] != lpPlan->lpszRangeHigh[i])? lpPlan->lpszRangeLow[i]: "", -1),
      SafeJavaScriptString(&safeStrings, &safeStringsLen, &safeStringsAlloc, (lpPlan->lpszRangeLow[i] != lpPlan->lpszRangeHigh[i])? lpPlan->lpszRangeHigh[i]: "", -1),
      (lpPlan->bRangeLowInclusive[i])? TRUE: FALSE, (lpPlan->bRangeHighInclusive[i])? TRUE: FALSE);

    /* Reset safeStringsLen */
    safeStringsLen = origSafeStringsLen;
  }
  printf("\n");

  /* Counter */
  printf("  // Counter\n");
  printf("  this.counterBegin = %u;\n", lpPlan->dwCounterBegin);
  printf("  this.counterIncrement = %u;\n", lpPlan->dwCounterIncrement);
  printf("  this.counterPrev = %u;\n", lpPlan->dwCounterPrev);
  printf("  this.counterNext = %u;\n", lpPlan->dwCounterNext);
  printf("  this.counterHints = %u;\n", (lpPlan->bCounterHints)? 1: 0);
  printf("  this.counterDisplayed = %u;\n", lpPlan->dwCounterDisplayed);
  printf("  this.counterTotal = %u;\n\n", lpPlan->dwCounterTotal);

  /* Background color */
  printf("  // Background color\n");
  printf("  this.useBackgroundColor = %u;\n", lpPlan->bUseBackgroundColor);
  if (lpPlan->bUseBackgroundColor)
    printf("  this.backgroundColor = \"%06x\";\n\n", lpPlan->dwBackgroundColor);
  else
    printf("\n");

  /* Sequence number */
  printf("  // Sequence number\n");
  printf("  this.sequenceNumber = %u;\n\n", lpPlan->dwSequenceNumber);

  /* JavaScript output */
  printf("  // JavaScript output\n");
  printf("  this.javaScript = 1;\n\n");

  /* Version information */
  printf("  // Version information\n");
  printf("  this.version = \"DirList2 %i.%i\";\n", DIRLIST_VERSION_MAJOR, DIRLIST_VERSION_MINOR);
  printf("  this.majorVersion = %i;\n", DIRLIST_VERSION_MAJOR);
  printf("  this.minorVersion = %i;\n\n", DIRLIST_VERSION_MINOR);

  /* Method declarations */
  printf("  // Method declarations\n");
  printf("  this.toString = DirList2ToString;\n");
  printf("}\n");
  printf("DirList2Included = true;\n");

  /* If we have a script footer, then write it now */
  if (lpszScriptFooter)
    printf("%s", lpszScriptFooter);

  /* End the JavaScript escape */
  printf("// --></script>\n");
}

void WriteHTMLHeader(LPCTSTR listName, LPCTSTR style) {
  FILE *header;
  int hHeader;
  UCHAR data[8192]; /* read 8k at a time */
  DWORD bytesRead;
  LPTSTR fileName = malloc(27 + strlen(listName) + strlen(style));
  sprintf(fileName, "dirlist_html/%s.%s.header.html", listName, style);
  hHeader = open(fileName, O_RDONLY, S_IREAD);
  free(fileName);
  header = fdopen(hHeader, "r");
  if (header) {
    /* Read until the end, writing everything to stdout */
    do {
      bytesRead = fread(data, 1, 8192, header);
      fwrite(data, 1, bytesRead, stdout);
    } while (bytesRead == 8192);

    /* Close the file */
    fclose(header);
    close(hHeader);
    return;
  }
  if (hHeader > 0)
    close(hHeader);
}

void WriteHTMLFooter(LPCTSTR listName, LPCTSTR style) {
  FILE *footer;
  int hFooter;
  UCHAR data[8192]; /* read 8k at a time */
  DWORD bytesRead;
  LPTSTR fileName = malloc(27 + strlen(listName) + strlen(style));
  sprintf(fileName, "dirlist_html/%s.%s.footer.html", listName, style);
  hFooter = open(fileName, O_RDONLY, S_IREAD);
  free(fileName);
  footer = fdopen(hFooter, "r");
  if (footer) {
    /* Read until the end, writing everything to stdout */
    do {
      bytesRead = fread(data, 1, 8192, footer);
      fwrite(data, 1, bytesRead, stdout);
    } while (bytesRead == 8192);

    /* Close the file */
    fclose(footer);
    close(hFooter);
    return;
  }
  if (hFooter > 0)
    close(hFooter);
}

BOOL ListDirectory(LPCTSTR lpszIndexName, LPDIRLIST_ACCESSPLAN lpPlan, UCHAR cDelimiter, UCHAR cStringChar, BOOL bJavaScript, LPDIRLIST_HEADER *lpHeader) {
  FILE *index;
  int hIndex;
  char bLoggedIn = FALSE;

  /* Set the guest credentials if no user was given or if user is Administrator ("root squashing," I suppose ;) */
  if (!lpPlan->lpUser) {
    lpPlan->lpUser = (LPDIRLIST_USER) malloc(sizeof(DIRLIST_USER));
    memset(lpPlan->lpUser, 0, sizeof(DIRLIST_USER));
    lpPlan->lpUser->lpszName = strdup("guest");
    lpPlan->lpUser->lpszPassword = strdup("");
  } else if (!lpPlan->lpUser->lpszName[0] || (CompareStrings(lpPlan->lpUser->lpszName, "Administrator") == 0)) {
    free(lpPlan->lpUser->lpszName);
    if (lpPlan->lpUser->lpszPassword)
      free(lpPlan->lpUser->lpszPassword);
    lpPlan->lpUser->lpszName = strdup("guest");
    lpPlan->lpUser->lpszPassword = strdup("");
  }

  /* Load the source file into memory, if any */
  *lpHeader = NULL;
  hIndex = open(lpszIndexName, O_RDONLY, S_IREAD);
  if (hIndex >= 0)
  {
    /* Check the file header */
    DWORD ver;
    index = fdopen(hIndex, "r");
    if (!(ver = CheckIndexHeader(index))) {
      infprintf("#Unknown file type or major version is too high.");
    } else
    {
      LPDIRLIST_ENTRY lpDirEntry;
      LPDIRLIST_CLASS lpClass = NULL;
      LPDIRLIST_CLASS lpPrevClass = NULL;
      BOOL verOk = TRUE;
      BOOL bSortFound = FALSE;
      LPBYTE lpData;
      int type;

      /* If this is a version 2 header, then read the header entry */
      if (ver > 2)
        verOk = FALSE;
      if (ver == 2) {
        if (ReadIndexEntry(index, &lpData, NULL) == 0)
        {
          /* Check the version information, ensuring we were meant to read this file */
          *lpHeader = (LPDIRLIST_HEADER) lpData;
          if ((*lpHeader)->dwLowestVersion > DIRLIST_VERSION_MINOR)
            verOk = FALSE;

          /* If encryption is set, then enable encryption */
          if ((*lpHeader)->dwFlags & HEADER_ENCRYPTFILE)
            bEncryptEnable = TRUE;
          else
            bEncryptEnable = FALSE;
        } else
        {
          /* Not a valid version 2 index file (the @0 is _required_)! */
          verOk = FALSE;
        }
      }

      if (verOk) {
        LPDIRLIST_SORTTABLE lpTable = NULL;
        BOOL bRaiseObsoleteWarning = FALSE;
        BOOL bSpinComplete, bUseSpin;
        DWORD dwCounter = 0;

        /* Determine if we need to use the spin */
        if (lpPlan->dwCounterTotal)
          bUseSpin = TRUE;
        else
          bUseSpin = FALSE;

        /* If no increment exists, then reset the other counter variables */
        if (!lpPlan->dwCounterIncrement) {
          lpPlan->dwCounterBegin = 1;
          lpPlan->dwCounterPrev = 0;
          lpPlan->dwCounterNext = 0;
          lpPlan->bCounterHints = FALSE;
        }

        /* Check for counter obsolescence */
        if ((*lpHeader)->dwSequenceNumber != lpPlan->dwSequenceNumber)
        {
          /* If the beginning range is not zero or one, then process further */
          if (lpPlan->dwCounterBegin > 1)
          {
            /* Reset the counter positions */
            lpPlan->dwCounterPrev = 0;
            lpPlan->dwCounterBegin = 0;

            /* Set the flag to remind us to raise a warning via JavaScript later */
            bRaiseObsoleteWarning = TRUE;

            /* We also need to recalculate the counter total, if any */
            lpPlan->dwCounterTotal = 0;
            bUseSpin = FALSE;
          }
        }

        /* If the beginning position was not set, then set it */
        if (!lpPlan->dwCounterBegin)
          lpPlan->dwCounterBegin = 1;
        lpPlan->dwCounterDisplayed = 0;

        /* Loop through the file */
        bSpinComplete = FALSE;
        type = ReadIndexEntry(index, &lpData, lpClass);
        while (!bSortFound && (type != -2) && (type != -3) && (!lpPlan->dwCounterIncrement || (!(bSpinComplete && bUseSpin) || (lpPlan->dwCounterDisplayed < lpPlan->dwCounterIncrement))))
        {
          switch (type) {

          case 1:
          case 2:

            /* Directory list entry -- display it proudly :-)! */
            /* (security checking is performed within DisplayPerson) */
            lpDirEntry = (LPDIRLIST_ENTRY) lpData;
            if (bLoggedIn && CheckRelevance(lpDirEntry, lpClass, lpPlan)) {
              if (!bUseSpin)
                (lpPlan->dwCounterTotal)++;
              if (lpPlan->dwCounterIncrement && (lpPlan->dwCounterDisplayed >= lpPlan->dwCounterIncrement))
                bSpinComplete = TRUE;
              else {
                dwCounter++;
                if (dwCounter >= lpPlan->dwCounterBegin) {
                  DisplayPerson(lpDirEntry, lpPlan, lpClass, cDelimiter, cStringChar);
                  (lpPlan->dwCounterDisplayed)++;
                }
              }
            }

            /* Free the memory associated with the directory entry */
            FreeEntry(lpDirEntry);
            free(lpDirEntry);
            break;

          case 3:

            /* Class information entry -- add it to the list :-)! */
            ((LPDIRLIST_CLASS) lpData)->next = NULL;
            if (lpPrevClass)
              lpPrevClass->next = (LPBYTE) lpData;
            else
              lpClass = (LPDIRLIST_CLASS) lpData;
            lpPrevClass = (LPDIRLIST_CLASS) lpData;
            break;

          case 4:

            /* Sort table entry -- if this is ours, read the data and set bSortFound */
            if (((LPDIRLIST_SORTTABLE) lpData)->idSort == lpPlan->idSort) {
              lpTable = (LPDIRLIST_SORTTABLE) lpData;
              bSortFound = TRUE;
            } else
              FreeSortTables((LPDIRLIST_SORTTABLE) lpData);
            break;

          case 5:

            /* Check to see if this user is valid if we're not logged in */
            if (!bLoggedIn)
            {
              /* If the current user is the user in the plan, then authenticate */
              if (CompareStrings(((LPDIRLIST_USER) lpData)->lpszName, lpPlan->lpUser->lpszName) == 0) {
                if (LoginUser((LPDIRLIST_USER *) &lpData, lpPlan, OPERATION_WEBQUERY, FALSE)) {
                  bLoggedIn = TRUE;

                  /* Use the loaded user structure as the user in the access plan */
                  FreeUsers(lpPlan->lpUser);
                  lpPlan->lpUser = (LPDIRLIST_USER) lpData;
                  lpPlan->lpUser->next = NULL;

                  /* Enable auditing if getting entries is an auditable event for this user */
                  bAuditEnable = OPERATION_GETENTRIES & lpPlan->lpUser->dwAuditOperations;
                } else
                {
                  /* Display the "not specific enough" error if guest; login error otherwise */
                  if (!bJavaScript || cDelimiter)
                    infprintf("#");
                  if (CompareStrings(lpPlan->lpUser->lpszName, "guest") == 0)
                    infprintf("Search not specific enough; follow the instructions on the <a href=\"javascript:history.go(-1);\">previous page</a> and try again.");
                  else
                    infprintf("Login unsuccessful due to bad username or password and/or insufficient permissions.");
                  FreeUsers((LPDIRLIST_USER) lpData);
                  if (*lpHeader) {
                    FreeHeader(*lpHeader);
                    free(*lpHeader);
                    *lpHeader = NULL;
                  }
                  if (lpClass)
                    FreeClasses(lpClass);
                  fclose(index);
                  close(hIndex);
                  return FALSE;
                }
              }
            }

            /* Free this user if we're not logged in */
            if (!bLoggedIn)
              FreeUsers((LPDIRLIST_USER) lpData);
          }

          /* Next entry */
          if (!bSortFound)
            type = ReadIndexEntry(index, &lpData, lpClass);
        }

        /* If the counter was changed, then hints cannot be used (not 100% sort-table based) */
        if (dwCounter)
          lpPlan->bCounterHints = FALSE;

        /* If we came to the end of the file, then set dwCounterNext to zero */
        if (type == -2)
          lpPlan->dwCounterNext = 0;
        if (lpPlan->dwCounterDisplayed + lpPlan->dwCounterBegin > lpPlan->dwCounterTotal)
          lpPlan->dwCounterNext = 0;
        else
          lpPlan->dwCounterNext = dwCounter + 1;

        /* If this user was not logged in, give an error */
        if (!bLoggedIn) {
          infprintf("-Could not authenticate because user does not exist or was squashed during login.");
          if (*lpHeader) {
            FreeHeader(*lpHeader);
            free(*lpHeader);
            *lpHeader = NULL;
          }
          if (lpClass)
            FreeClasses(lpClass);
          fclose(index);
          close(hIndex);
          return FALSE;
        }

        /* If we're using a found sort table, display the entries in order */
        if (bSortFound)
        {
          /* Read the sort table into an array and then go through it :-) */
          LPDWORD lpdwTableEntries = ReadSortEntriesArray(index, lpTable);
          DWORD i = 0;
          if (lpdwTableEntries)
          {
            /* Seek to the starting position in the sort table (set to zero if out of range) */
            DWORD dwIDecrease = 0;
            if (lpPlan->dwCounterBegin && lpPlan->bCounterHints) {
              if (lpPlan->dwCounterBegin <= lpTable->nEntries) {
                dwCounter = lpPlan->dwCounterBegin - 1;
                i = lpPlan->dwCounterBegin - 1;
              } else
                lpPlan->dwCounterBegin = 1;
            }

            /* Process each sort table entry */
            bSpinComplete = FALSE;
            type = 0;
            while ((i < lpTable->nEntries) && (type != -3) && (type != -2) && (!lpPlan->dwCounterIncrement || (!(bSpinComplete && bUseSpin) || (lpPlan->dwCounterDisplayed < lpPlan->dwCounterIncrement))))
            {
              /* Go to the file position of the index entry */
              SeekFile(index, (long) lpdwTableEntries[i], SEEK_SET);

              /* Read the index entry */
              type = ReadIndexEntry(index, &lpData, lpClass);

              /* If this is a type-1 or type-2 entry (should be), then print it :-) */
              if ((type == 1) || (type == 2)) {
                lpDirEntry = (LPDIRLIST_ENTRY) lpData;
                if (CheckRelevance(lpDirEntry, lpClass, lpPlan)) {
                  if (!bUseSpin)
                    (lpPlan->dwCounterTotal)++;
                  if (lpPlan->dwCounterIncrement && (lpPlan->dwCounterDisplayed >= lpPlan->dwCounterIncrement)) {
                    bSpinComplete = TRUE;
                    if (bUseSpin)
                      i--;
                  } else {
                    dwCounter++;
                    if (dwCounter >= lpPlan->dwCounterBegin) {
                      DisplayPerson(lpDirEntry, lpPlan, lpClass, cDelimiter, cStringChar);
                      (lpPlan->dwCounterDisplayed)++;
                    }
                  }
                }

                /* Free the memory associated with the directory entry */
                FreeEntry(lpDirEntry);
                free(lpDirEntry);
              }

              /* If we're just counting now, then increase the decrease variable */
              if (bSpinComplete && !bUseSpin)
                dwIDecrease++;

              /* Next sort table entry */
              i++;
            }

            /* Decrease i by dwIDecrease */
            i -= dwIDecrease;

            /* If the client wanted counter hints instead of relative positions, then do it */
            if (lpPlan->bCounterHints)
              lpPlan->dwCounterNext = i + 1;
            else
              lpPlan->dwCounterNext = dwCounter + 1;

            /* If we hit the end of the list, then always reset dwCounterNext */
            if (i >= lpTable->nEntries)
              lpPlan->dwCounterNext = 0;

            /* Free the sort table array */
            free(lpdwTableEntries);
          } else
          {
            /* Error!  Sort requested, but sort table data could not be read! */
            infprintf("#Error!  Sort table data area could not be read!\n");
            fflush(stdout);
          }
        }

        /* Set the sequence number of the plan equal to the actual file's number */
        lpPlan->dwSequenceNumber = (*lpHeader)->dwSequenceNumber;

        /* Write the JavaScript stuff to the output if we're supposed to */
        if (bJavaScript && !cDelimiter)
        {
          /* If we need to raise an obsolescence warning, then set the script footer */
          LPTSTR szScriptFooter = NULL;
          if (bRaiseObsoleteWarning)
            szScriptFooter = "alert(\"The index was refreshed and made your previous search results obsolete.\\n\\nThis listing contains your new search results, starting from the beginning.\");\n";

          /* Write the access plan as a JavaScript :-) */
          WriteAccessPlanAsJavaScript(lpPlan, szScriptFooter);
        } else
        {
          /* Because we're not using JavaScript, we'll need to just write the obsolescence */
          /* warning if one is set */
          if (bRaiseObsoleteWarning)
            infprintf("#Sequence number obsolete -- started from beginning with new data\n");
        }

        /* If we're supposed to have found it but it doesn't exist, tell the user */
        if (!bSortFound && lpPlan->idSort)
          infprintf("#Sort table %u not found in the database!", lpPlan->idSort);

        /* Free class information and the sort tables */
        FreeClasses(lpClass);
        FreeSortTables(lpTable);
      } else
      {
        /* Index file cannot be read by us due to versioning -- tell the user */
        if (*lpHeader)
          infprintf("#Program version not high enough -- upgrade to read this file (File Major: %u, File Minor: %u, File Minimum Minor: %u, Program Major: %u, Program Minor %u)\n", ver, (*lpHeader)->dwVersion, (*lpHeader)->dwLowestVersion, DIRLIST_VERSION_MAJOR, DIRLIST_VERSION_MINOR);
        else
          infprintf("#Program version not high enough -- upgrade to read this file (File Major: %u, Program Major: %u, Program Minor %u)\n", ver, DIRLIST_VERSION_MAJOR, DIRLIST_VERSION_MINOR);
      }
    }

    /* Close the index */
    fclose(index);
    close(hIndex);

    /* We're done */
    return TRUE;
  } else {
    infprintf("#Could not obtain read access to the index \"%s\" (it might not exist).\n", lpszIndexName);
    return FALSE;
  }
}

BOOL GetAccessPlan(LPTSTR *lpCGIVars, FILE *pInput, LPDWORD lpdwMaxRead, UCHAR cDelimiter, LPDIRLIST_ACCESSPLAN lpPlan, BOOL bSearches, BOOL bCounters, BOOL bEmitStatus, LPTSTR *lpszError)
{
  /* Clear the access plan's memory first (ensures all defaults are NULL) */
  memset(lpPlan, 0, sizeof(DIRLIST_ACCESSPLAN));

  /* Get the value filters for display */
  lpPlan->dwFields = strtoul(GetQuery(lpCGIVars, "f", "0"), NULL, 10);
  lpPlan->dwXFields = strtoul(GetQuery(lpCGIVars, "xf", "0"), NULL, 10);
  lpPlan->dwExtFields = strtoul(GetQuery(lpCGIVars, "ef", "0"), NULL, 10);
  lpPlan->dwXExtFields = strtoul(GetQuery(lpCGIVars, "xef", "0"), NULL, 10);
  lpPlan->dwExtValues = strtoul(GetQuery(lpCGIVars, "ev", "0"), NULL, 10);
  lpPlan->dwExtBoth = strtoul(GetQuery(lpCGIVars, "eb", "0"), NULL, 10);
  lpPlan->dwExtFlags = strtoul(GetQuery(lpCGIVars, "efl", "0"), NULL, 10);

  /* Get the background color */
  lpPlan->bUseBackgroundColor = (strtoul(GetQuery(lpCGIVars, "ubg", "0"), NULL, 16))? TRUE: FALSE;
  lpPlan->dwBackgroundColor = strtoul(GetQuery(lpCGIVars, "bg", "ffffff"), NULL, 16);

  /* Get the login information, if any */
  lpPlan->lpUser = NULL;
  if (strtoul(GetQuery(lpCGIVars, "a", "0"), NULL, 10))
  {
    /* Read the authentication string */
    LPTSTR lpszLine = (LPTSTR) malloc(32);
    LPTSTR lpszCurrent;
    DWORD dwLineLen = 0, dwLineAlloc = 32;
    if (ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead))
    {
      /* Allocate a user structure and fill it out */
      lpPlan->lpUser = (LPDIRLIST_USER) malloc(sizeof(DIRLIST_USER));
      memset(lpPlan->lpUser, 0, sizeof(DIRLIST_USER));

      /* Get the username and password */
      lpszCurrent = strtok2(lpszLine, cDelimiter);
      if (lpszCurrent) {
        lpPlan->lpUser->lpszName = strdup(lpszCurrent);
        lpszCurrent = strtok2(NULL, cDelimiter);
        if (lpszCurrent)
          lpPlan->lpUser->lpszPassword = strdup(lpszCurrent);
        else {
          if (bEmitStatus)
            *lpszError = strdup("#Authentication enabled but no password was given.\n");
          free(lpPlan->lpUser->lpszName);
          free(lpPlan->lpUser);
          free(lpszLine);
          lpPlan->lpUser = NULL;
          return FALSE;
        }
      } else {
        if (bEmitStatus)
          *lpszError = strdup("#Authentication enabled but no login was given.\n");
        free(lpPlan->lpUser);
        free(lpszLine);
        lpPlan->lpUser = NULL;
        return FALSE;
      }
    }

    /* Free the line */
    free(lpszLine);
  }

  /* Get the administrative mode */
  lpPlan->bAdminMode = (strtoul(GetQuery(lpCGIVars, "am", "0"), NULL, 10))? TRUE: FALSE;

  /* Get the sort table ID */
  lpPlan->idSort = strtoul(GetQuery(lpCGIVars, "o", "0"), NULL, 10);

  /* Get the flag modification types */
  lpPlan->wFlagModifyType = (WORD) strtoul(GetQuery(lpCGIVars, "fm", "0"), NULL, 10);
  lpPlan->wExtFlagModifyType = (WORD) strtoul(GetQuery(lpCGIVars, "efm", "0"), NULL, 10);

  /* Get the requested sequence number */
  lpPlan->dwSequenceNumber = strtoul(GetQuery(lpCGIVars, "seq", "0"), NULL, 10);

  /* Get the search/range parameters if requested */
  if (bSearches)
  {
    lpPlan->nSearches = strtoul(GetQuery(lpCGIVars, "ns", "0"), NULL, 10);
    if (lpPlan->nSearches)
    {
      /* If we have a number of searches that would likely exhaust the memory and/or CPU,
         then don't run it */
      DWORD i;
      if (lpPlan->nSearches > QUERY_MAXSEARCH) {
        if (lpPlan->lpUser != NULL) {
          free(lpPlan->lpUser->lpszName);
          free(lpPlan->lpUser->lpszPassword);
          free(lpPlan->lpUser);
          lpPlan->lpUser = NULL;
        }
        *lpszError = strdup("#Too many searches.\n");
        return FALSE;
      }

      /* Allocate the memory for these entries */
      lpPlan->dwSearchFields = malloc(lpPlan->nSearches * sizeof(DWORD));
      lpPlan->dwExtSearchFields = malloc(lpPlan->nSearches * sizeof(DWORD));
      lpPlan->dwExtSearchOrdinals = malloc(lpPlan->nSearches * sizeof(DWORD));
      lpPlan->bSearchNot = malloc(lpPlan->nSearches * sizeof(BOOL));
      lpPlan->bSearchAnd = malloc(lpPlan->nSearches * sizeof(BOOL));
      lpPlan->dwSearchLevel = malloc(lpPlan->nSearches * sizeof(DWORD));
      lpPlan->dwSearchLevelEnd = malloc(lpPlan->nSearches * sizeof(DWORD));
      lpPlan->fSearchExact = malloc(lpPlan->nSearches * sizeof(DWORD));
      lpPlan->lpszSearch = malloc(lpPlan->nSearches * sizeof(LPTSTR));
      lpPlan->lpszRangeLow = malloc(lpPlan->nSearches * sizeof(LPTSTR));
      lpPlan->lpszRangeHigh = malloc(lpPlan->nSearches * sizeof(LPTSTR));
      lpPlan->bRangeLowInclusive = malloc(lpPlan->nSearches * sizeof(BOOL));
      lpPlan->bRangeHighInclusive = malloc(lpPlan->nSearches * sizeof(BOOL));

      /* If the last allocation didn't work, then write out our failure and return FALSE */
      if (lpPlan->bRangeHighInclusive == NULL) {
        if (lpPlan->dwSearchFields != NULL) free(lpPlan->dwSearchFields);
        if (lpPlan->dwExtSearchFields != NULL) free(lpPlan->dwExtSearchFields);
        if (lpPlan->dwExtSearchOrdinals != NULL) free(lpPlan->dwExtSearchOrdinals);
        if (lpPlan->bSearchNot != NULL) free(lpPlan->bSearchNot);
        if (lpPlan->bSearchAnd != NULL) free(lpPlan->bSearchAnd);
        if (lpPlan->dwSearchLevel != NULL) free(lpPlan->dwSearchLevel);
        if (lpPlan->dwSearchLevelEnd != NULL) free(lpPlan->dwSearchLevelEnd);
        if (lpPlan->fSearchExact != NULL) free(lpPlan->fSearchExact);
        if (lpPlan->lpszSearch != NULL) free(lpPlan->lpszSearch);
        if (lpPlan->lpszRangeLow != NULL) free(lpPlan->lpszRangeLow);
        if (lpPlan->lpszRangeHigh != NULL) free(lpPlan->lpszRangeHigh);
        if (lpPlan->bRangeLowInclusive != NULL) free(lpPlan->bRangeLowInclusive);
        if (lpPlan->lpUser != NULL) {
          free(lpPlan->lpUser->lpszName);
          free(lpPlan->lpUser->lpszPassword);
          free(lpPlan->lpUser);
          lpPlan->lpUser = NULL;
        }
        *lpszError = strdup("#Not enough memory available to run query.\n");
        return FALSE;
      }

      /* Gather the search/range parameters */
      for (i = 0; i < lpPlan->nSearches; i++) {
        lpPlan->dwSearchFields[i] = strtoul(GetQueryArray(lpCGIVars, "sf", i, "0"), NULL, 10);
        lpPlan->dwExtSearchFields[i] = strtoul(GetQueryArray(lpCGIVars, "esf", i, "0"), NULL, 10);
        lpPlan->dwExtSearchOrdinals[i] = strtoul(GetQueryArray(lpCGIVars, "eso", i, "0"), NULL, 10);
        lpPlan->bSearchNot[i] = (strtoul(GetQueryArray(lpCGIVars, "sn", i, "0"), NULL, 10))? TRUE: FALSE;
        lpPlan->bSearchAnd[i] = (strtoul(GetQueryArray(lpCGIVars, "sa", i, "0"), NULL, 10))? TRUE: FALSE;
        lpPlan->dwSearchLevel[i] = strtoul(GetQueryArray(lpCGIVars, "sl", i, "0"), NULL, 10);
        lpPlan->dwSearchLevelEnd[i] = strtoul(GetQueryArray(lpCGIVars, "sle", i, "0"), NULL, 10);
        lpPlan->fSearchExact[i] = strtoul(GetQueryArray(lpCGIVars, "se", i, "0"), NULL, 10);
        lpPlan->lpszSearch[i] = TrimString(strdup(GetQueryArray(lpCGIVars, "s", i, "")));
        lpPlan->lpszRangeLow[i] = GetQueryArray(lpCGIVars, "rl", i, "");
        lpPlan->lpszRangeHigh[i] = GetQueryArray(lpCGIVars, "rh", i, "");
        lpPlan->bRangeLowInclusive[i] = (strtoul(GetQueryArray(lpCGIVars, "rli", i, "0"), NULL, 10))? TRUE: FALSE;
        lpPlan->bRangeHighInclusive[i] = (strtoul(GetQueryArray(lpCGIVars, "rhi", i, "0"), NULL, 10))? TRUE: FALSE;

        /* If fSearchExact is SEARCH_LIKE, then set the escape to backslash if not set */
        if (lpPlan->fSearchExact[i] & SEARCH_LIKE) {
          if (!((lpPlan->fSearchExact[i] & SEARCH_ESCAPEMASK) / 256))
            lpPlan->fSearchExact[i] |= (DWORD) '\\' * 256;
        }

        /* Give the strings unique memory */
        if (lpPlan->lpszSearch[i][0] != 0) {
          lpPlan->lpszRangeLow[i] = TrimString(strdup(lpPlan->lpszRangeLow[i]));
          lpPlan->lpszRangeHigh[i] = lpPlan->lpszRangeLow[i];
        } else {
          lpPlan->lpszRangeLow[i] = TrimString(strdup(lpPlan->lpszRangeLow[i]));
          lpPlan->lpszRangeHigh[i] = TrimString(strdup(lpPlan->lpszRangeHigh[i]));
        }

        /* If we need to set the search back, then do so :-) */
        if ((lpPlan->lpszSearch[i][0] == 0) && (lpPlan->lpszRangeLow[i][0] != 0)) {
          free(lpPlan->lpszSearch[i]);
          lpPlan->lpszSearch[i] = strdup(lpPlan->lpszRangeLow[i]);
        }
      }
    }
  }

  /* Get the counter values if requested */
  if (bCounters)
  {
    lpPlan->dwCounterBegin = strtoul(GetQuery(lpCGIVars, "cb", "0"), NULL, 10);
    lpPlan->dwCounterIncrement = strtoul(GetQuery(lpCGIVars, "ci", "0"), NULL, 10);
    lpPlan->dwCounterPrev = strtoul(GetQuery(lpCGIVars, "cp", "0"), NULL, 10);
    lpPlan->dwCounterNext = strtoul(GetQuery(lpCGIVars, "cn", "0"), NULL, 10);
    lpPlan->dwCounterTotal = strtoul(GetQuery(lpCGIVars, "ct", "0"), NULL, 10);
    lpPlan->bCounterHints = (strtoul(GetQuery(lpCGIVars, "ch", "0"), NULL, 10))? TRUE: FALSE;
  }

  /* We did it :-)! */
  return TRUE;
}

/* Write the access plan as a log format into output */
void WriteAccessPlanLog(FILE *output, LPDIRLIST_ACCESSPLAN lpPlan)
{
  /* Start the log message */
  DWORD i;
  fprintf(output, "Access Plan from Client:\n");

  /* Authentication */
  if (lpPlan->bAdminMode) fprintf(output, "  adminMode = %i\n", lpPlan->bAdminMode);

  /* Value filters */
  if (lpPlan->dwFields) fprintf(output, "  fields = %u\n", lpPlan->dwFields);
  if (lpPlan->dwXFields) fprintf(output, "  xFields = %u\n", lpPlan->dwXFields);
  if (lpPlan->dwExtFields) fprintf(output, "  extFields = %u\n", lpPlan->dwExtFields);
  if (lpPlan->dwXExtFields) fprintf(output, "  xExtFields = %u\n", lpPlan->dwXExtFields);
  if (lpPlan->dwExtValues) fprintf(output, "  extValues = %u\n", lpPlan->dwExtValues);
  if (lpPlan->dwExtBoth) fprintf(output, "  extBoth = %u\n", lpPlan->dwExtBoth);
  if (lpPlan->dwExtFlags) fprintf(output, "  extFlags = %u\n", lpPlan->dwExtFlags);
  if (lpPlan->idSort) fprintf(output, "  sort = %u\n", lpPlan->idSort);
  if (lpPlan->nSearches) fprintf(output, "  nSearches = %u\n", lpPlan->nSearches);
  for (i = 0; i < lpPlan->nSearches; i++)
  {
    /* Write out the details of searches that actually have details */
    if (lpPlan->dwSearchFields[i] || lpPlan->dwExtSearchFields[i] || lpPlan->dwExtSearchOrdinals[i] || lpPlan->bSearchNot[i] ||
        lpPlan->bSearchAnd[i] || lpPlan->dwSearchLevel[i] || lpPlan->dwSearchLevelEnd[i] ||
        lpPlan->fSearchExact[i] || lpPlan->lpszRangeLow[i][0] || lpPlan->lpszRangeHigh[i][0] || lpPlan->lpszSearch[i][0] ||
        lpPlan->bRangeLowInclusive[i] || lpPlan->bRangeHighInclusive[i])
    {
      /* Write out this search's details */
      fprintf(output, "  search %u:\n", i);
      if (lpPlan->dwSearchFields[i]) fprintf(output, "    searchFields = %u\n", lpPlan->dwSearchFields[i]);
      if (lpPlan->dwExtSearchFields[i]) fprintf(output, "    extSearchFields = %u\n", lpPlan->dwExtSearchFields[i]);
      if (lpPlan->dwExtSearchOrdinals[i]) fprintf(output, "    extSearchOrdinals = %u\n", lpPlan->dwExtSearchOrdinals[i]);
      if (lpPlan->bSearchNot[i]) fprintf(output, "    searchNot = %i\n", lpPlan->bSearchNot[i]);
      if (lpPlan->bSearchAnd[i]) fprintf(output, "    searchAnd = %i\n", lpPlan->bSearchAnd[i]);
      if (lpPlan->dwSearchLevel[i]) fprintf(output, "    searchLevel = %u\n", lpPlan->dwSearchLevel[i]);
      if (lpPlan->dwSearchLevelEnd[i]) fprintf(output, "    searchLevelEnd = %u\n", lpPlan->dwSearchLevelEnd[i]);
      if (lpPlan->fSearchExact[i]) fprintf(output, "    searchExact = %u\n", lpPlan->fSearchExact[i]);
      if (lpPlan->lpszRangeLow[i] == lpPlan->lpszRangeHigh[i]) fprintf(output, "    search = %s\n", lpPlan->lpszSearch[i]);
      if (lpPlan->lpszRangeLow[i] != lpPlan->lpszRangeHigh[i]) {
        fprintf(output, "    rangeLow = %s\n", lpPlan->lpszRangeLow[i]);
        fprintf(output, "    rangeHigh = %s\n", lpPlan->lpszRangeHigh[i]);
        fprintf(output, "    rangeLowInclusive = %i\n", lpPlan->bRangeLowInclusive[i]);
        fprintf(output, "    rangeHighInclusive = %i\n", lpPlan->bRangeHighInclusive[i]);
      } else {
        if (lpPlan->bRangeLowInclusive[i]) fprintf(output, "    rangeLowInclusive = %i\n", lpPlan->bRangeLowInclusive[i]);
        if (lpPlan->bRangeHighInclusive[i]) fprintf(output, "    rangeHighInclusive = %i\n", lpPlan->bRangeHighInclusive[i]);
      }
    }
  }

  /* Counters */
  if (lpPlan->dwCounterBegin) fprintf(output, "  counterBegin = %u\n", lpPlan->dwCounterBegin);
  if (lpPlan->dwCounterIncrement) fprintf(output, "  counterIncrement = %u\n", lpPlan->dwCounterIncrement);
  if (lpPlan->dwCounterPrev) fprintf(output, "  counterPrev = %u\n", lpPlan->dwCounterPrev);
  if (lpPlan->dwCounterNext) fprintf(output, "  counterNext = %u\n", lpPlan->dwCounterNext);
  if (lpPlan->bCounterHints) fprintf(output, "  counterHints = %u\n", lpPlan->bCounterHints);
  if (lpPlan->dwCounterDisplayed) fprintf(output, "  counterDisplayed = %u\n", lpPlan->dwCounterDisplayed);
  if (lpPlan->dwCounterTotal) fprintf(output, "  counterTotal = %u\n", lpPlan->dwCounterTotal);

  /* Background color */
  if (lpPlan->bUseBackgroundColor) {
    fprintf(output, "  useBackgroundColor = %u\n", lpPlan->bUseBackgroundColor);
    fprintf(output, "  backgroundColor = #%06x\n", lpPlan->dwBackgroundColor);
  }

  /* Sequence number */
  if (lpPlan->dwSequenceNumber) fprintf(output, "  sequenceNumber = %u\n", lpPlan->dwSequenceNumber);

  /* End the log message */
  fprintf(output, "\n");
}

/*************************************************************************/
/**                                                                     **/
/**     getcgivars.c-- routine to read CGI input variables into an      **/
/**         array of strings.                                           **/
/**                                                                     **/
/**     The x2c() and unescape_url() routines were lifted directly      **/
/**     from NCSA's sample program util.c, packaged with their HTTPD.   **/
/**                                                                     **/
/*************************************************************************/

/* Convert a two-char hex string into the char it represents */
char x2c(LPTSTR what) {
  register UCHAR digit;

  digit = (what[0] >= 'A' ? ((what[0] & 0xdf) - 'A')+10 : (what[0] - '0'));
  digit *= 16;
  digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A')+10 : (what[1] - '0'));
  return(digit);
}

/* Reduce any %xx escape sequences to the characters they represent */
void unescape_url(LPTSTR url) {
  register int i,j;

  for(i=0,j=0; url[j]; ++i,++j) {
    if((url[i] = url[j]) == '%') {
      url[i] = x2c(&url[j+1]) ;
      j+= 2 ;
    }
  }
  url[i] = '\0' ;
}

/* Read the CGI input and place all name/val pairs into list.        */
/* Returns list containing name1, value1, name2, value2, ... , NULL  */
char **getcgivars() {
  register int i ;
  LPTSTR request_method ;
  int content_length;
  LPTSTR cgiinput ;
  LPTSTR *cgivars ;
  LPTSTR *pairlist ;
  int paircount ;
  LPTSTR nvpair ;
  LPTSTR eqpos ;

  /* Depending on the request method, read all CGI input into cgiinput */
  /* (really should produce HTML error messages, instead of exit()ing) */
  request_method= getenv("REQUEST_METHOD") ;
  if (!request_method)
    return NULL;
  if (!strcmp(request_method, "GET")) {
    cgiinput = getenv("QUERY_STRING");
    if (!cgiinput)
      cgiinput = strdup("");
    else
      cgiinput = strdup(cgiinput);
  }
  else if (!strcmp(request_method, "POST")) {
    /* strcasecmp() is not supported in Windows-- use strcmpi() instead */
    if ( strcasecmp(getenv("CONTENT_TYPE"), "application/x-www-form-urlencoded")) {
    /*      printf("getcgivars(): Unsupported Content-Type.\n") ;
      exit(1) ;*/
      cgiinput = getenv("QUERY_STRING");
      if (!cgiinput)
        cgiinput = strdup("");
      else
        cgiinput = strdup(cgiinput);
      if ( !(content_length = atoi(getenv("CONTENT_LENGTH"))) ) {
        printf("getcgivars(): No Content-Length was sent with the POST request.\n") ;
        exit(1) ;
      }
    } else {
      if ( !(content_length = atoi(getenv("CONTENT_LENGTH"))) ) {
        printf("getcgivars(): No Content-Length was sent with the POST request.\n") ;
        exit(1) ;
      }
      if ( !(cgiinput= (LPTSTR) malloc(content_length+1)) ) {
        printf("getcgivars(): Could not malloc for cgiinput.\n") ;
        exit(1) ;
      }
      if (!ReadFile(cgiinput, content_length, 1, stdin)) {
        printf("Couldn't read CGI input from STDIN.\n") ;
        exit(1) ;
      }
      cgiinput[content_length]='\0' ;
    }
  }
  else {
    printf("getcgivars(): unsupported REQUEST_METHOD\n") ;
    exit(1) ;
  }

  /* Change all plusses back to spaces */
  for(i=0; cgiinput[i]; i++) if(cgiinput[i] == '+') cgiinput[i] = ' ' ;

  /* First, split on "&" to extract the name-value pairs into pairlist */
  pairlist= (LPTSTR *) malloc(256*sizeof(LPTSTR *)) ;
  paircount= 0 ;
  nvpair= strtok(cgiinput, "&") ;
  while (nvpair) {
    pairlist[paircount++]= strdup(nvpair) ;
    if (!(paircount%256))
      pairlist= (LPTSTR *) realloc(pairlist,(paircount+256)*sizeof(LPTSTR *)) ;
    nvpair= strtok(NULL, "&") ;
  }
  pairlist[paircount]= 0 ;    /* terminate the list with NULL */

  /* Then, from the list of pairs, extract the names and values */
  cgivars= (LPTSTR *) malloc((paircount*2+1)*sizeof(LPTSTR *)) ;
  for (i= 0; i<paircount; i++) {
    if (eqpos=strchr(pairlist[i], '=')) {
      *eqpos= '\0' ;
      unescape_url(cgivars[i*2+1]= strdup(eqpos+1)) ;
    } else {
      unescape_url(cgivars[i*2+1]= strdup("")) ;
    }
    unescape_url(cgivars[i*2]= strdup(pairlist[i])) ;
  }
  cgivars[paircount*2]= 0 ;   /* terminate the list with NULL */

  /* Free anything that needs to be freed */
  free(cgiinput) ;
  for (i=0; pairlist[i]; i++) free(pairlist[i]) ;
  free(pairlist) ;

  /* Return the list of name-value strings */
  return cgivars ;
}

/* **************** end of the getcgivars() module ******************* */

/* Frees all the CGI variables */
void FreeCGIVars(LPTSTR *lpCGIVars)
{
  /* Free every string variable until no more are found */
  int i;
  for (i = 0; lpCGIVars[i]; i++)
    free(lpCGIVars[i]);

  /* Free the array itself */
  free(lpCGIVars);
}

/* The following was added by Steven Lawrance */
LPTSTR GetQuery(LPTSTR *lpCGIVars, LPTSTR lpszParamName, LPTSTR lpszDefaultValue)
{
  /* Get a variable */
  BOOL found = FALSE;
  int i = 0;
  if (!lpCGIVars)
    return lpszDefaultValue;
  while (!found && (lpCGIVars[i] != NULL)) {
    if (CompareStrings(lpCGIVars[i], lpszParamName) == 0)
      found = TRUE;
    else
      i += 2;
  }

  /* If we found it, then return its value */
  if (found)
    return lpCGIVars[i + 1];
  else
    return lpszDefaultValue;
}

LPTSTR GetQueryArray(LPTSTR *lpCGIVars, LPTSTR lpszParamName, DWORD dwIndex, LPTSTR lpszDefaultValue)
{
  /* Get a variable */
  LPTSTR lpszIndexName = malloc(strlen(lpszParamName) + 11);
  BOOL found = FALSE;
  int i = 0;
  sprintf(lpszIndexName, "%s%u", lpszParamName, dwIndex);
  if (!lpCGIVars)
    return lpszDefaultValue;
  while (!found && (lpCGIVars[i] != NULL)) {
    if (CompareStrings(lpCGIVars[i], lpszIndexName) == 0)
      found = TRUE;
    else
      i += 2;
  }
  free(lpszIndexName);

  /* If we found it, then return its value */
  if (found)
    return lpCGIVars[i + 1];
  else
    return lpszDefaultValue;
}
